#
# This file contains the methods used to estimate the effect of a treatment in
# an outcome, given a graph.
#
from typing import Tuple

import networkx as nx
import pandas as pd
from dowhy import CausalModel
from dowhy.causal_estimator import CausalEstimate
from dowhy.causal_refuter import CausalRefuter
from sklearn.preprocessing import StandardScaler

from causalexplain.common import utils


def estimate(
        digraph: nx.DiGraph, 
        data: pd.DataFrame, 
        in_place: bool = True, 
        verbose=False) -> nx.DiGraph:
    """
    Takes the original digraph passed as argument, and computes the effect of
    each treatment in the outcome, given the graph. Treament and outcome are
    the pairs formed by traversing the edges in the oriented DAG.
    The resulting ATE and RefuteResult are computed by calling the method
    `estimate_edge` and incorporated to each edge by adding the ATE and
    RefuteResult as attributes.

    Parameters
    ----------
    digraph : nx.DiGraph
        The causal graph.
    data : pd.DataFrame
        The data.
    in_place : bool, optional
        Whether to modify the graph in place. Defaults to True.
    verbose : bool, optional
        If True, print the results. Defaults to False.

    Returns
    -------
    nx.DiGraph
        The estimated causal graph.
    """
    # Create a copy of the graph
    graph = digraph if in_place else digraph.copy()

    # Compute the effect of each treatment in the outcome
    for t, o in graph.edges():
        ate, refute_pval = estimate_edge(graph, t, o, data, verbose)
        graph.add_edge(t, o, ate=ate, refute_pval=refute_pval)

    print(graph.edges(data=True)) if verbose else None

    return graph


def estimate_edge(
        digraph: nx.DiGraph,
        treatment: str,
        outcome: str,
        data: pd.DataFrame,
        verbose=False) -> Tuple[CausalEstimate, CausalRefuter]:
    """
    Estimate the effect of a treatment in an outcome, given a graph.

    Parameters
    ----------
    digraph : nx.DiGraph
        The causal graph.
    treatment : str
        The name of the treatment variable.
    outcome : str
        The name of the outcome variable.
    data : pd.DataFrame
        The data.
    verbose : bool, optional
        If True, print the results. Defaults to False.

    Returns
    -------
    CausalEstimate
        The estimated effect.
    """
    # Definir el modelo causal
    model = CausalModel(
        data=data,
        treatment=treatment,
        outcome=outcome,
        graph=digraph
    )

    # Identificar el efecto causal
    identified_estimand = model.identify_effect(
        proceed_when_unidentifiable=True)
    print(identified_estimand) if verbose else None

    # Estimar el efecto causal
    estimate = model.estimate_effect(
        identified_estimand,
        target_units="ate",
        method_name="backdoor.linear_regression"
    )
    print(estimate) if verbose else None

    # Pruebas de refutación
    refute = model.refute_estimate(
        identified_estimand,
        estimate,
        method_name="random_common_cause"
    )
    print(refute) if verbose else None
    return (estimate.value, refute.refutation_result['p_value'])


def main(exp_name,
         data_path="/Users/renero/phd/data/RC3/",
         output_path="/Users/renero/phd/output/RC4/",
         scale=False):
    """
    Runs a custom main function for the given experiment name.

    Args:
        experiment_name (str): The name of the experiment to run.
        path (str): The path to the data files.
        output_path (str): The path to the output files.

    Returns:
        None
    """

    ref_graph = utils.graph_from_dot_file(f"{data_path}{exp_name}.dot")
    data: pd.DataFrame = pd.read_csv(f"{data_path}{exp_name}.csv")
    if scale:
        scaler = StandardScaler()
        data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)
    # Split the dataframe into train and test

    rex = utils.load_experiment(f"{exp_name}_nn", output_path)
    print(f"Loaded experiment {exp_name}")

    n_dag = estimate(rex.G_shap, data, verbose=True)


if __name__ == "__main__":
    main("rex_generated_linear_0")



import importlib
import pathlib
import sys
from typing import Iterable

import networkx as nx
import numpy as np
import pytest

sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

sid_module = importlib.import_module("causalexplain.metrics.SID")


@pytest.fixture(autouse=True)
def _matplotlib_cache(tmp_path, monkeypatch):
    monkeypatch.setenv("MPLCONFIGDIR", str(tmp_path))


@pytest.fixture(autouse=True)
def _networkx_matrix_aliases(monkeypatch):
    """NetworkX>=3 removed matrix helpers; reintroduce them so SID utilities work unchanged."""
    if not hasattr(nx, "from_numpy_matrix"):
        monkeypatch.setattr(
            nx, "from_numpy_matrix", nx.from_numpy_array, raising=False
        )
    if not hasattr(nx, "to_numpy_matrix"):
        monkeypatch.setattr(
            nx,
            "to_numpy_matrix",
            lambda graph, **kwargs: np.matrix(nx.to_numpy_array(graph, **kwargs)),
            raising=False,
        )


@pytest.fixture
def restore_debug_flag():
    original = sid_module.DEBUG
    yield
    sid_module.DEBUG = original


def test_debug_prints_message_when_flag_enabled(capsys, restore_debug_flag):
    sid_module.DEBUG = True
    sid_module.debug_("hello", "world")
    captured = capsys.readouterr().out
    assert "hello world" in captured


def test_pm_outputs_matrix_rows(capsys, restore_debug_flag):
    sid_module.DEBUG = True
    sid_module.pm_(np.array([[1, 0], [0, 2]]))
    output = capsys.readouterr().out.splitlines()
    assert output[0].startswith("00")
    assert output[-1].strip().endswith("2")


def test_all_dags_intern_returns_column_major_flatten():
    gm = np.matrix([[1, 2], [3, 4]])
    a = np.zeros((2, 2))
    result = sid_module.allDagsIntern(gm, a, np.array([0, 1]))
    assert np.array_equal(result, np.array([[1, 3, 2, 4]]))


def test_all_dags_intern_raises_on_partially_directed_input():
    gm = np.identity(2)
    a = np.array([[0, 1], [0, 0]])
    with pytest.raises(ValueError):
        sid_module.allDagsIntern(gm, a, np.array([0, 1]))


def test_all_dags_jonas_returns_negative_one_for_directed_component():
    adj = np.matrix([[0, 1], [0, 0]])
    assert sid_module.allDagsJonas(adj, np.array([0, 1])) == -1


def test_all_dags_jonas_defers_to_all_dags_intern_for_trivial_component():
    adj = np.matrix([[0, 0], [0, 0]])
    result = sid_module.allDagsJonas(adj, np.array([0]))
    assert np.array_equal(result, np.array([[0, 0, 0, 0]]))


@pytest.mark.parametrize("sparse", [False, True])
def test_compute_path_matrix_matches_transitive_closure(sparse):
    adjacency = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]])
    matrix = sid_module.computePathMatrix(adjacency, spars=sparse)
    matrix = matrix.toarray() if sparse else matrix
    expected = np.array([[1, 1, 1], [0, 1, 1], [0, 0, 1]], dtype=bool)
    assert np.array_equal(matrix, expected)


def test_compute_path_matrix_two_blocks_with_conditioning():
    """Conditioning on a node should zero-out downstream reachability in PathMatrix2."""
    adjacency = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]], dtype=int)
    base = sid_module.computePathMatrix(adjacency)
    conditioned = sid_module.computePathMatrix2(adjacency.copy(), [1], base)
    assert not bool(conditioned[0, 2])
    assert base[0, 2]


def test_compute_caus_order_prefers_sources_first():
    adjacency = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]], dtype=float)
    assert sid_module.compute_caus_order(adjacency) == [1, 2, 3]


def test_dag2cpdag_adj_identity_for_chain_graph(monkeypatch):
    """dag2cpdagAdj should preserve a fully directed chain."""
    monkeypatch.setattr(
        sid_module,
        "compute_caus_order",
        lambda adj: list(range(adj.shape[0])),
    )
    monkeypatch.setattr(sid_module, "dag2cpdag", lambda graph: graph)
    adjacency = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]], dtype=float)
    cpdag = sid_module.dag2cpdagAdj(adjacency)
    expected = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]], dtype=float)
    assert np.array_equal(cpdag, expected)


def test_dag2cpdag_adj_returns_original_when_empty():
    adjacency = np.zeros((2, 2))
    assert sid_module.dag2cpdagAdj(adjacency) is adjacency


def test_dsep_adji_reports_reachability_from_source():
    """dSepAdji should mark descendants of a node reachable when nothing is conditioned."""
    adjacency = np.matrix([[0, 1, 0], [0, 0, 1], [0, 0, 0]])
    cond_set: Iterable[int] = []
    path = sid_module.computePathMatrix(adjacency.copy())
    pruned = sid_module.computePathMatrix2(adjacency.copy(), cond_set, path)
    result = sid_module.dSepAdji(
        adjacency.copy(),
        i=0,
        condSet=np.array(cond_set, dtype=int),
        PathMatrix=path,
        PathMatrix2=pruned,
        spars=False,
        p=adjacency.shape[1],
    )
    assert np.array_equal(result["reachableJ"], np.array([True, True, False]))
    assert not result["reachableOnNonCausalPath"].any()


def test_unique_rows_returns_first_occurrence_indices():
    matrix = np.array([[1, 0], [1, 0], [0, 1], [0, 1]])
    assert sid_module.unique_rows(matrix) == [0, 2]


def test_get_ind_in_all_others_identifies_matching_parent_sets():
    mmm = np.array([[1, 0, 1], [1, 0, 1], [0, 1, 0]])
    unique = np.array([0])
    parents = np.array([0, 2])
    matches = sid_module.get_indInAllOthers(2, mmm, unique, parents, 1, np.array([1, 2]))
    assert np.array_equal(matches, np.array([0]))


def test_get_ind_in_all_others_returns_empty_when_no_match():
    mmm = np.array([[1, 0, 1], [0, 1, 0]])
    result = sid_module.get_indInAllOthers(
        2,
        mmm,
        np.array([0]),
        np.array([0, 2]),
        1,
        np.array([1]),
    )
    assert result.size == 0


def test_hamming_distance_switches_formulas():
    g1 = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]])
    g2 = np.array([[0, 1, 0], [0, 0, 0], [0, 1, 0]])
    assert sid_module.hammingDist(g1, g1) == 0
    assert sid_module.hammingDist(g1, g2, allMistakesOne=True) == 1
    assert sid_module.hammingDist(g1, g2, allMistakesOne=False) == 2


def test_sid_sparse_branch_surfaces_sparse_limitations():
    """When requesting sparse paths the current implementation raises during np.isnan checks."""
    graph = np.array([[0, 1], [0, 0]])
    with pytest.raises(TypeError):
        sid_module.SID(graph, graph, spars=True)


# Legacy SID consistency tests (formerly in test_metrics_SID_legacy.py)


def _is_dag(adj_matrix: np.ndarray) -> bool:
    graph = nx.from_numpy_array(adj_matrix, create_using=nx.DiGraph)
    return nx.is_directed_acyclic_graph(graph)


def _compared_sid(trueGraph: np.ndarray, estGraph: np.ndarray, sid_fn) -> dict:
    if not _is_dag(estGraph):
        return {"sid": 0, "sidLowerBound": 0.0, "sidUpperBound": 0.0}
    if not isinstance(trueGraph, np.ndarray):
        trueGraph = np.array(trueGraph)
    if not isinstance(estGraph, np.ndarray):
        estGraph = np.array(estGraph)
    trueGraph = np.matrix(trueGraph)
    estGraph = np.matrix(estGraph)
    s = sid_fn(trueGraph, estGraph, edge_direction="from row to column")
    return {"sid": s[1], "sidLowerBound": s[0], "sidUpperBound": s[0]}


def _random_dag(size: int, probability: float) -> np.ndarray:
    rng = np.random.default_rng(0)
    adj = rng.binomial(1, probability, size=(size, size)).astype(np.int8)
    adj = np.triu(adj, 1)
    perm = rng.permutation(size)
    return adj[perm, :][:, perm]


@pytest.fixture
def gadjid_sid():
    mod = pytest.importorskip("gadjid", reason="gadjid package required for legacy SID tests")
    return mod.sid


def test_manual_sid_matches_expected(gadjid_sid):
    G = np.array(
        [
            [0, 1, 1, 1, 1],
            [0, 0, 1, 1, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    H1 = np.array(
        [
            [0, 1, 1, 1, 1],
            [0, 0, 1, 1, 1],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    H2 = np.array(
        [
            [0, 0, 1, 1, 1],
            [1, 0, 1, 1, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    H3 = np.array(
        [
            [0, 1, 1, 1, 1],
            [1, 0, 1, 1, 1],
            [1, 1, 0, 1, 0],
            [1, 1, 1, 0, 0],
            [1, 1, 0, 0, 0],
        ]
    )
    H4 = np.array(
        [
            [0, 0, 1, 1, 1],
            [0, 0, 0, 1, 0],
            [1, 1, 0, 1, 0],
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
        ]
    )
    H5 = np.array(
        [
            [0, 0, 1, 1, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )

    sid1 = sid_module.SID(G, H1)
    assert sid1["sid"] == 0
    sid2 = sid_module.SID(G, H2)
    assert sid2["sid"] == 8
    sid3 = sid_module.SID(G, H3)
    assert sid3["sidLowerBound"] == 0
    assert sid3["sidUpperBound"] == 15
    sid4 = sid_module.SID(G, H4)
    assert sid4["sidLowerBound"] == 8
    assert sid4["sidUpperBound"] == 16
    sid5 = sid_module.SID(G, H5)
    assert sid5["sid"] == 12


def test_compared_sid_consistent_with_sid(gadjid_sid):
    n = 20
    threshold = 0.2
    count = 0
    for _ in range(n):
        p = np.random.randint(3, 20)
        G = _random_dag(p, 0.2)
        H = G.copy()
        indices = np.where(G == 1)
        for i in range(len(indices[0])):
            if np.random.random() > threshold:
                H[indices[0][i]][indices[1][i]] = 0
        sid1 = sid_module.SID(G, H, output=False)
        sid2 = _compared_sid(G, H, sid_fn=gadjid_sid)
        if sid1["sid"] == sid2["sid"]:
            count += 1
    assert count / n == 1.0



import numpy as np
import pandas as pd
import networkx as nx
import pytest

from causalexplain.generators import generators
from causalexplain.generators.mechanisms import LinearMechanism


@pytest.fixture(autouse=True)
def _seed():
    np.random.seed(0)


@pytest.fixture
def deterministic_graph(monkeypatch):
    """Force adjacency sampling to succeed by deterministically selecting edges."""

    def fake_randint(low, high=None, size=None):
        if high is None:
            high = low
            low = 0
        return low + 1 if (high - low) > 1 else low

    def fake_choice(seq, size, replace=False):
        seq = list(seq)
        size = int(size)
        if size <= 0:
            return np.array([], dtype=int)
        return np.array(seq[-size:])

    monkeypatch.setattr(generators.np.random, "randint", fake_randint)
    monkeypatch.setattr(generators.np.random, "choice", fake_choice)


def test_init_variables_creates_mechanisms(deterministic_graph):
    gen = generators.AcyclicGraphGenerator(
        causal_mechanism="linear", nodes=3, parents_max=2, points=10
    )
    gen.init_variables()
    assert isinstance(gen.g, nx.DiGraph)
    # Leaf node should use linear mechanism, roots keep initial generator callable
    assert any(isinstance(func, LinearMechanism) for func in gen.cfunctions)


def test_generate_initializes_on_demand(monkeypatch):
    gen = generators.AcyclicGraphGenerator("linear", nodes=3, points=5)

    def fake_init():
        gen.adjacency_matrix = np.zeros((gen.nodes, gen.nodes))
        gen.g = nx.DiGraph(gen.adjacency_matrix)
        gen.cfunctions = [lambda n, verbose=False: np.arange(n) for _ in range(gen.nodes)]

    monkeypatch.setattr(gen, "init_variables", fake_init)
    graph, data = gen.generate(rescale=False)
    assert graph.number_of_nodes() == 3
    assert (data.iloc[:, 0] == np.arange(5)).all()


def test_generate_rescales_columns(deterministic_graph):
    gen = generators.AcyclicGraphGenerator("linear", nodes=3, parents_max=2, points=30)
    gen.init_variables()
    _, data = gen.generate(rescale=True)
    assert pytest.approx(data.iloc[:, 0].mean(), abs=1e-6) == 0


def test_to_csv_requires_generated_data(tmp_path):
    gen = generators.AcyclicGraphGenerator("linear")
    gen.data = None
    with pytest.raises(ValueError):
        gen.to_csv(str(tmp_path / "sample"))


def test_to_csv_writes_all_artifacts(tmp_path):
    gen = generators.AcyclicGraphGenerator("linear", nodes=2, points=3)
    gen.data = pd.DataFrame({"V0": [0, 1], "V1": [1, 2]})
    gen.adjacency_matrix = np.array([[0, 1], [0, 0]])
    gen.g = nx.DiGraph(gen.adjacency_matrix)
    gen.to_csv(str(tmp_path / "dataset"))
    assert (tmp_path / "dataset_data.csv").exists()
    assert (tmp_path / "dataset_target.csv").exists()
    assert (tmp_path / "dataset_target.dot").exists()



import numpy as np
import pytest

from causalexplain.estimators.cam.computeScoreMatParallel import computeScoreMatParallel


def test_compute_score_mat_parallel_semgam_runs(monkeypatch):
    monkeypatch.setattr(
        "causalexplain.estimators.cam.computeScoreMatParallel.train_gam",
        lambda X, y, pars=None, verbose=False: {"residuals": np.array([1.0, 3.0])},
    )
    row_parents = np.array([[0], [1]])
    sel_mat = np.ones((2, 2), dtype=bool)
    X = np.array([[0.0, 1.0], [1.0, 2.0]])

    score = computeScoreMatParallel(
        row_parents=row_parents,
        score_name="SEMGAM",
        X=X,
        sel_mat=sel_mat,
        verbose=False,
        node2=1,
        i=0,
        pars_score={},
        interv_mat=None,
        interv_data=False,
    )

    assert np.isclose(score, 0.0)


def test_compute_score_mat_parallel_prunes_unselected_parent(monkeypatch):
    row_parents = np.array([[0], [1]])
    sel_mat = np.array([[True, False], [True, True]])
    X = np.ones((2, 2))

    score = computeScoreMatParallel(
        row_parents=row_parents,
        score_name="SEMGAM",
        X=X,
        sel_mat=sel_mat,
        verbose=False,
        node2=1,
        i=0,
        pars_score={},
        interv_mat=None,
        interv_data=False,
    )

    assert score == -np.inf


@pytest.mark.parametrize(
    "score_name,exc_type",
    [("SEMSEV", ValueError), ("SEMIND", NotImplementedError)],
)
def test_compute_score_mat_parallel_raises_for_invalid_scores(score_name, exc_type):
    row_parents = np.array([[0], [1]])
    sel_mat = np.ones((2, 2), dtype=bool)
    X = np.ones((2, 2))

    with pytest.raises(exc_type):
        computeScoreMatParallel(
            row_parents=row_parents,
            score_name=score_name,
            X=X,
            sel_mat=sel_mat,
            verbose=False,
            node2=1,
            i=0,
            pars_score={},
            interv_mat=None,
            interv_data=False,
        )



import numpy as np
import networkx as nx
import pandas as pd
import pytest
import torch

from causalexplain.models import _weight_utils as wu


class DummyParam:
    def __init__(self, array):
        self._array = array

    def detach(self):
        return self

    def numpy(self):
        return self._array


class DummyModule:
    def __init__(self):
        self._weights = [
            ("weight", DummyParam(np.array([[1.0, -1.0], [0.5, 0.25]]))),
            ("bias", DummyParam(np.array([0.0, 0.1]))),
        ]

    def named_parameters(self):
        return self._weights


def test_extract_weights_filters_non_weights():
    weights = wu.extract_weights(DummyModule())
    assert len(weights) == 1
    assert weights[0].shape == (2, 2)


def test_plot_helpers_no_display(monkeypatch):
    monkeypatch.setattr(wu.plt, "show", lambda *_, **__: None)
    W = np.array([[0.1, 0.2], [0.3, -0.1]])
    wu.see_weights_to_hidden(W, ["f1", "f2"], target="y")
    wu.see_weights_from_input(W, ["f1", "f2"], target="y")


def test_plot_feature_composes_from_weights(monkeypatch):
    monkeypatch.setattr(wu.plt, "show", lambda *_, **__: None)
    dummy = DummyModule()

    class Result:
        def __init__(self, model):
            self.model = model
            self.columns = ["f1", "Noise"]
            self.all_columns = ["f1", "target", "Noise"]

    res = Result(dummy)
    wu.plot_feature(res)
    wu.plot_features({"f1": res, "target": res, "Noise": res}, 3, 3, res.all_columns)


def test_layer_and_summarize_weights(monkeypatch):
    class ExtendedModule(DummyModule):
        def __init__(self):
            self._weights = [
                ("weight", DummyParam(np.array([[1.0, -1.0, 0.5], [0.5, 0.25, -0.25]]))),
                ("bias", DummyParam(np.array([0.0, 0.1, -0.1]))),
            ]

    dummy = ExtendedModule()
    monkeypatch.setattr(
        wu, "extract_weights", lambda model: list(model._weights[0][1]._array[None, ...])
    )

    class Wrapper:
        def __init__(self, model):
            self.model = model
            self.columns = ["f1", "f2", "Noise"]

    weights_map = {"f1": Wrapper(dummy), "f2": Wrapper(dummy)}
    layer_df = wu.layer_weights(weights_map["f1"], target="f1")
    assert {col for col in layer_df.columns} == {"psd_f1", "avg_f1", "med_f1"}

    summary = wu.summarize_weights(weights_map, feature_names=["f1", "f2"], scale=False)
    assert summary.shape[0] == dummy._weights[0][1]._array.shape[1]


def test_summarize_weights_scales_and_plots(monkeypatch):
    monkeypatch.setattr(wu.plt, "show", lambda *_, **__: None)
    values = np.array([[1.0, 0.5], [0.2, -0.4]])

    class Wrapper:
        def __init__(self):
            self.model = type("M", (), {"named_parameters": lambda *_: [("weight", DummyParam(values))]})
            self.columns = ["a", "Noise"]

    weights_map = {"a": Wrapper(), "b": Wrapper()}
    scaled = wu.summarize_weights(weights_map, feature_names=["a", "b"], scale=True)
    assert np.allclose(scaled.mean().values, 0, atol=1e-6)

    weights_df = pd.DataFrame(
        [[0.0, 0.0, 0.0, 0.1, 0.1, 0.1], [1.0, 1.0, 1.0, 0.5, 0.5, 0.5]],
        columns=["psd_a", "med_a", "avg_a", "psd_b", "med_b", "avg_b"],
        index=["a", "b"],
    )
    relationships = wu.identify_relationships(
        weights_df, feature_names=["a", "b"], plot=True, eps=0.5, min_counts=1
    )
    assert set(relationships.keys()) == {"a", "b"}


def test_identify_relationships_detects_sparse_clusters():
    columns = [
        "psd_a",
        "med_a",
        "avg_a",
        "psd_b",
        "med_b",
        "avg_b",
        "psd_c",
        "med_c",
        "avg_c",
    ]
    values = [
        [0.1, 0.1, 0.2, 0.5, 0.5, 0.6, 1.0, 1.1, 1.2],
        [2.0, 1.5, 1.0, 2.5, 2.0, 2.1, 2.2, 2.3, 2.4],
        [4.0, 3.5, 3.0, 3.5, 3.0, 3.1, 3.2, 3.3, 3.4],
    ]
    data = pd.DataFrame(values, columns=columns, index=["a", "b", "c"])

    rels = wu.identify_relationships(data, feature_names=["a", "b", "c"], plot=False, eps=0.1, min_counts=1)
    assert set(rels.keys()) == {"a", "b", "c"}


def test_shap_value_helpers(monkeypatch):
    class FakeDeep:
        def __init__(self, model, data):
            self.model = model
            self.data = data

        def shap_values(self, values):
            return np.array(values)

    class DummyLoader:
        def __init__(self, features):
            self.dataset = type("DS", (), {"features": features})

    class DummyModel:
        def __init__(self):
            self.model = object()
            self.train_loader = DummyLoader(np.array([[1.0, 2.0], [3.0, 4.0]]))

    monkeypatch.setattr(wu.shap, "DeepExplainer", FakeDeep)

    dummy_model = DummyModel()
    shap_values = wu._get_shap_values(dummy_model)
    assert shap_values.shape == dummy_model.train_loader.dataset.features.shape

    avg = wu._average_shap_values({"col": shap_values}, ["col"], abs=True)
    assert avg.shape[0] == 1
    signed_avg = wu._average_shap_values({"col": shap_values}, ["col"], abs=False)
    assert signed_avg.shape == avg.shape


def test_find_shap_elbow_uses_knee_locator(monkeypatch):
    class FakeKnee:
        def __init__(self, *_, **__):
            self.elbow = 1

        def plot_knee(self, *_, **__):
            return None

    monkeypatch.setattr(wu.kneed, "KneeLocator", FakeKnee)
    values = np.array([0.1, 0.2, 0.3, 0.4])
    threshold = wu._find_shap_elbow(values, plot=True, verbose=True)
    assert threshold >= 0


def test_identify_edges_and_orientation():
    avg_shaps = np.array([[0.9], [0.8]])
    threshold = {"a": 0.5, "b": 0.5}
    edges = wu._identify_edges(avg_shaps, ["a", "b"], threshold)
    assert ("b", pytest.approx(avg_shaps[0][0])) in edges["a"]

    g = nx.DiGraph()
    g.add_edge("a", "b", weight=1.0)
    g.add_edge("b", "a", weight=0.8)
    oriented = wu._orient_edges_based_on_shap(g, verbose=True)
    assert oriented.has_edge("a", "b") != oriented.has_edge("b", "a")


def test_remove_asymmetric_shap_edges():
    relations = {"a": [("b", 1.0)], "b": []}
    cleaned = wu._remove_asymmetric_shap_edges(relations)
    assert cleaned["a"] == []
    assert cleaned["b"] == []


def test_infer_causal_relationships(monkeypatch):
    class FakeProgBar:
        def start_subtask(self, *args, **kwargs):
            return self

        def update_subtask(self, *_, **__):
            return None

        def remove(self, *_, **__):
            return None

    class FakeKnee:
        def __init__(self, *_, **__):
            self.elbow = 1

        def plot_knee(self, *_, **__):
            return None

    class FakeDeep:
        def __init__(self, model, data):
            self.model = model
            self.data = data

        def shap_values(self, values):
            return np.array(values)

    class DummyLoader:
        def __init__(self, features):
            self.dataset = type("DS", (), {"features": features})

    class DummyModel:
        def __init__(self, name):
            self.model = object()
            data = np.array([[1.0, 2.0], [2.0, 1.0]])
            self.train_loader = DummyLoader(data)
            self.name = name

    monkeypatch.setattr(wu.shap, "DeepExplainer", FakeDeep)
    monkeypatch.setattr(wu.kneed, "KneeLocator", FakeKnee)
    monkeypatch.setattr(wu, "ProgBar", FakeProgBar)
    monkeypatch.setattr(
        wu,
        "graph_from_dictionary",
        lambda d: (
            lambda g: (
                [g.add_edge(u, v, weight=w) for u, rels in d.items() for v, w in rels],
                g,
            )[1]
        )(nx.DiGraph()),
    )

    models = {name: DummyModel(name) for name in ["a", "b"]}
    result = wu.infer_causal_relationships(
        trained_models=models,
        feature_names=list(models.keys()),
        prune=True,
        plot=False,
        prog_bar=True,
    )

    assert "graph" in result
    assert isinstance(result["graph"], nx.DiGraph)



import pytest

from causalexplain.estimators.ges import decomposable_score


def test_local_score_caches_results(monkeypatch):
    calls = {"compute": 0}

    class Dummy(decomposable_score.DecomposableScore):
        def _compute_local_score(self, x, pa):
            calls["compute"] += 1
            return x + len(pa)

    score = Dummy(data=None, cache=True, debug=0)

    first = score.local_score(1, {2, 3})
    second = score.local_score(1, {3, 2})  # same set different order hits cache
    assert first == 3
    assert second == 3
    assert calls["compute"] == 1


def test_local_score_without_cache_calls_each_time():
    class Dummy(decomposable_score.DecomposableScore):
        def _compute_local_score(self, x, pa):
            return x + len(pa)

    score = Dummy(data=None, cache=False)
    assert score.local_score(0, set()) == 0
    assert score.local_score(0, set()) == 0



import pytest
import networkx as nx

from causalexplain.common.utils import correct_edge_from_prior

class TestCorrectEdgeFromPrior:

    """
        a ---> b
       ^ \     |
      /   v    v
      c-->d    e
          |   ^
          v  /
          f
    """
    dag = nx.DiGraph()
    dag.add_edge('a', 'b')
    dag.add_edge('c', 'a')
    dag.add_edge('a', 'd')
    dag.add_edge('c', 'd')
    dag.add_edge('b', 'e')
    dag.add_edge('d', 'f')
    dag.add_edge('f', 'e')
    prior = [['a', 'b'], ['c', 'd', 'e'], ['f']]


    def test_both_in_top_list(self):
        assert correct_edge_from_prior(self.dag, 'a', 'b', self.prior, False) ==  1
        assert not self.dag.has_edge('a', 'b')

    def test_v_is_before_u(self):
        # These edges violate the prior order.
        assert correct_edge_from_prior(self.dag, 'c', 'a', self.prior, True) == -1
        assert correct_edge_from_prior(self.dag, 'f', 'e', self.prior, True) == -1
        # Check that 'dag' is changed
        assert not self.dag.has_edge('a', 'c')
        assert not self.dag.has_edge('f', 'e')

    def test_u_is_before_v(self):
        # These edges are in the correct order.
        assert correct_edge_from_prior(self.dag, 'a', 'd', self.prior, False) == 1
        assert correct_edge_from_prior(self.dag, 'd', 'f', self.prior, False) == 1
        assert correct_edge_from_prior(self.dag, 'b', 'e', self.prior, False) == 1

    def test_not_in_clear_order(self):
        assert correct_edge_from_prior(self.dag, 'c', 'd', self.prior, False) == 0

    def test_edge_not_present(self):
        assert correct_edge_from_prior(self.dag, 'a', 'h', self.prior, False) == 0



import numpy as np
import scipy
import scipy.optimize
import networkx as nx


def run(variant, data, loss, loss_grad, **kwargs):
    return variant(data, loss, loss_grad, **kwargs)


def notears_standard(
        data, loss, loss_grad, c=0.25, r=10.0, e=1e-8, rnd_W_init=False,
        output_all_progress=False, verbose=False, max_iter=None):
    """
    Runs NOTEARS algorithm.

    Args:
        data (np.array): n x d data matrix with n samples, d variables
        c (float): minimum rate of progress, c \in (0,1)
        r (float): penalty growth rate, r > 1
        e (float): optimation accuracy, e > 0 (acyclicity stopping criteria)
        loss (function): loss function
        loss_grad (function): gradient of the loss function
        rnd_W_init (bool): initialize W to std. normal random matrix, rather than
            zero matrix
        output_all_progress (bool): return all intermediate values of W, rather than
        just the final value
        verbose (bool): print optimization information
    Returns:
        dict: { 'h': acyclicity of output,
                'loss': loss of output,
                'W': resulting optimized adjacency matrix}
    """
    n = np.shape(data)[0]
    d = np.shape(data)[1]

    data = np.array(data).astype(dtype=np.float64)
    cov = np.cov(data.T)

    if rnd_W_init:
        W = np.random.randn(d, d)
    else:
        W = np.zeros([d, d])  # initial guess
    W = W.astype(dtype=np.float64)
    a = 0.0    # initial guess
    p = 1.0    # initial penalty

    if output_all_progress:
        ret = []

    def h(W):
        # tr exp(W ◦ W) − d
        return np.trace(scipy.linalg.expm(np.multiply(W, W))) - d

    def h_grad(W):
        # ∇h(W) = [exp(W ◦ W)]^T ◦ 2W
        return np.multiply(np.transpose(scipy.linalg.expm(np.multiply(W, W))), 2.0 * W)

    def L(W, p, a):
        W = np.reshape(W, [d, d]).astype(dtype=np.float64)
        return loss(W, data, cov, d, n) + (p/2.0)*(h(W)**2) + a*(h(W))

    def L_grad(W, p, a):
        W = np.reshape(W, [d, d]).astype(dtype=np.float64)
        return np.reshape(
            loss_grad(W, data, cov, d, n) + h_grad(W)*(a + (p*h(W))),
            [d**2]).astype(dtype=np.float64)

    def get_W_star(p, W, a):
        W_flat = W.flatten()
        W_star = scipy.optimize.minimize(L, W_flat, args=(
            p, a), jac=L_grad, method='L-BFGS-B', options={'disp': False})
        W_star.x = W_star.x.reshape(W.shape).astype(dtype=np.float64)
        return W_star

    it = 0
    while True:
        it += 1
        W_star = get_W_star(p, W, a)
        W_star = W_star.x
        # W_star = W_star.reshape(get_W_star(p, W, a)['x'],
        #   [d, d]).astype(dtype=np.float64)
        h_W_star = h(W_star)
        if h(W) != 0.0:
            while h_W_star >= max(c * h(W), e):
                p = r*p
                W_star = np.reshape(get_W_star(p, W, a)['x'], [
                                    d, d]).astype(dtype=np.float64)
                h_W_star = h(W_star)
                if verbose:
                    print("Increasing p:\t p = {: .2e}\n\t\t h_W_star = {}".format(
                        p, h_W_star))
        if output_all_progress:
            ret.append({'h': h_W_star, 'loss': loss(
                W_star, data, cov, d, n), 'a': a, 'W': W_star})
        if h_W_star < e:
            if verbose:
                print("Done:\t\t h = {}\n\t\t loss = {}\nt\t\t a = {}".format(
                    h_W_star, loss(W_star, data, cov, d, n), a))
            if output_all_progress:
                return ret
            return {'h': h_W_star, 'loss': loss(W_star, data, cov, d, n), 'W': W_star}
            if verbose:
                print("Progress:\t h = {}\n\t\t loss = {}\n\t\t a = {}". format(
                    h_W_star, loss(W_star, data, cov, d, n), a))
        a = a + p*h_W_star
        W = W_star
        if max_iter is not None and it >= max_iter:
            # Stop early if iteration cap reached
            if output_all_progress:
                return ret
            return {'h': h_W_star, 'loss': loss(W_star, data, cov, d, n), 'W': W_star}



import pandas as pd
import pytest

from causalexplain.independence import edge_orientation


class DummyResult:
    def __init__(self, pvalue):
        self.pvalue = pvalue


class DummyHsic:
    def __init__(self, sequence):
        self.sequence = list(sequence)
        self.calls = []

    def test(self, x, y, reps=None):
        self.calls.append((x, y, reps))
        return DummyResult(self.sequence.pop(0))


def _patch_dependencies(monkeypatch, pvalues):
    dummy = DummyHsic(pvalues)
    monkeypatch.setattr(edge_orientation, "Hsic", lambda: dummy)
    monkeypatch.setattr(
        edge_orientation,
        "fit_and_get_residuals",
        lambda *args, **kwargs: args[1],  # pass through second array
    )
    return dummy


@pytest.mark.parametrize(
    "pvalues,expected",
    [
        ([0.2, 0.05], 1),
        ([0.01, 0.3], -1),
        ([0.1, 0.1], 0),
    ],
)
def test_get_edge_orientation_branches(monkeypatch, pvalues, expected):
    data = pd.DataFrame({"x": [0, 1, 2], "y": [2, 1, 0]})
    dummy = _patch_dependencies(monkeypatch, pvalues)

    result = edge_orientation.get_edge_orientation(
        data, "x", "y", iters=5, method="gpr", verbose=True
    )

    assert result == expected
    assert dummy.calls[0][2] == 5  # reps forwarded


def test_get_edge_orientation_invalid_method_propagates(monkeypatch):
    data = pd.DataFrame({"x": [0, 1], "y": [1, 0]})
    with pytest.raises(ValueError):
        edge_orientation.get_edge_orientation(data, "x", "y", method="bad")



"""Acyclic Graph Generator.

Generates a dataset out of an acyclic FCM.
Author : Olivier Goudet and Diviyan Kalainathan

Minor update: J. Renero (due to pandas changes to as_matrix()), and bug fix in
polynomial mechanism.
"""

import numpy as np
import pandas as pd
import networkx as nx

from sklearn.preprocessing import scale

from causalexplain.generators.mechanisms import *
from causalexplain.common.plot import dag2dot


class AcyclicGraphGenerator(object):
    """Generates a cross-sectional dataset out of a cyclic FCM."""

    def __init__(self, causal_mechanism,
                 initial_variable_generator=gmm_cause,
                 points=500, nodes=20, timesteps=0, parents_max=5, verbose=False):
        """
        :params:
        @param:initial_variable_generator(points): init variables of the graph
        @param:causal_mechanism(causes): generating causes in the graph to
            choose between: ['linear', 'polynomial', 'sigmoid_add',
            'sigmoid_mix', 'gp_add', 'gp_mix']
        """
        super(AcyclicGraphGenerator, self).__init__()
        self.mechanism = {'linear': LinearMechanism,
                          'polynomial': Polynomial_Mechanism,
                          'sigmoid_add': SigmoidAM_Mechanism,
                          'sigmoid_mix': SigmoidMix_Mechanism,
                          'gp_add': GaussianProcessAdd_Mechanism,
                          'gp_mix': GaussianProcessMix_Mechanism}[causal_mechanism]
        self.data = pd.DataFrame(
            None, columns=[f"V{i}" for i in range(nodes)])
        self.nodes = nodes
        if timesteps == 0:
            self.timesteps = np.inf
        else:
            self.timesteps = timesteps
        self.points = points
        self.adjacency_matrix = np.zeros((nodes, nodes))
        self.parents_max = parents_max
        self.initial_generator = initial_variable_generator
        self.cfunctions = None
        self.g = None
        self.verbose = verbose

    def init_variables(self):
        """Redefine the causes of the graph."""
        # Resetting adjacency matrix
        for i in range(self.nodes-1):
            for j in np.random.choice(range(i+1, self.nodes),
                                      np.random.randint(0, min([self.parents_max,
                                                                self.nodes-i])),
                                      replace=False):
                if i != j:
                    self.adjacency_matrix[i, j] = 1

        try:
            assert any([sum(self.adjacency_matrix[:, i]) ==
                        self.parents_max for i in range(self.nodes)])
            self.g = nx.DiGraph(self.adjacency_matrix)
            assert not list(nx.simple_cycles(self.g))

        except AssertionError:
            if self.verbose:
                print("Regenerating, graph non valid...")
            self.init_variables()

        # Mechanisms
        if self.verbose:
            print("Generating mechanisms...")
        self.cfunctions = [
            self.mechanism(int(sum(self.adjacency_matrix[:, i])), self.points,
                           verbose=self.verbose)
            if sum(self.adjacency_matrix[:, i]) else self.initial_generator
            for i in range(self.nodes)
        ]

    def generate(self, rescale=True) -> (nx.DiGraph, pd.DataFrame):
        """Generate data from an FCM containing cycles."""
        if self.cfunctions is None:
            if self.verbose:
                print("CFunctions empty, initializing...")
            self.init_variables()

        for i in nx.topological_sort(self.g):
            if self.verbose:
                print(f"Generating V{i}")
            # Root cause
            if not sum(self.adjacency_matrix[:, i]):
                if self.verbose:
                    print(f"  V{i} is a root cause")
                self.data[f'V{i}'] = self.cfunctions[i](
                    self.points, verbose=self.verbose)
            # Generating causes
            else:
                if self.verbose:
                    print(
                        f"  V{i} parents: {self.adjacency_matrix[:, i].nonzero()[0]}")
                column = self.adjacency_matrix[:, i].nonzero()[0]
                self.data[f'V{i}'] = self.cfunctions[i](
                    np.stack(self.data.iloc[:, column].values),
                    verbose=self.verbose)
            if rescale:
                self.data[f'V{i}'] = scale(np.stack(self.data[f'V{i}'].values))

        return self.g, self.data

    def to_csv(self, fname_radical, **kwargs):
        """
        Save data to the csv format by default, in two separate files.

        Optional keyword arguments can be passed to pandas.
        """
        if self.data is None:
            raise ValueError("Graph has not yet been generated. \
                              Use self.generate() to do so.")

        self.data.to_csv(fname_radical+'_data.csv', **kwargs)
        pd.DataFrame(self.adjacency_matrix).to_csv(
            fname_radical+'_target.csv', **kwargs)
        # Save also the DOT format
        graph_dot_format = dag2dot(self.g).to_string()
        graph_dot_format = f"strict {graph_dot_format[:-9]}\n}}"
        with open(fname_radical+'_target.dot', "w") as f:
            f.write(graph_dot_format)


if __name__ == '__main__':
    save = True

    # Seed
    np.random.seed(1342)

    # Generate data
    g = AcyclicGraphGenerator("polynomial", points=20,
                              nodes=10, timesteps=0, parents_max=3, verbose=False)
    graph, data = g.generate()

    # Cross check experiment
    # import matplotlib.pyplot as plt
    # print("V1")
    # print(data.V1.values)
    # plt.scatter(data.V1, data.V3)
    # plt.xlabel("V1"); plt.ylabel("V3")
    # plt.show()

    if save:
        # Save to file
        data.to_csv(
            "/Users/renero/phd/data/RC3/rex_generated_polynew_1.csv", index=False)
        # write to file
        graph_dot_format = dag2dot(graph).to_string()
        graph_dot_format = f"strict {graph_dot_format[:-9]}\n}}"
        with open("/Users/renero/phd/data/RC3/rex_generated_polynew_1.dot", "w") as f:
            f.write(graph_dot_format)



import networkx as nx
import numpy as np
import pytest

import causalexplain.estimators.pc.dag as dag_module
from causalexplain.estimators.pc.dag import DAG


def test_cycle_detection_raises():
    with pytest.raises(ValueError):
        DAG([("A", "B"), ("B", "A")])


def test_add_nodes_from_weight_mismatch():
    dag = DAG()
    with pytest.raises(ValueError):
        dag.add_nodes_from(["A", "B"], weights=[1])


def test_moralize_and_immoralities():
    dag = DAG([("A", "C"), ("B", "C")])
    immoralities = dag.get_immoralities()
    assert ("A", "B") in {tuple(sorted(x)) for x in immoralities}

    dag_module.UndirectedGraph = nx.Graph
    moral = dag.moralize()
    assert moral.has_edge("A", "B")


def test_getters_and_do_operation():
    dag = DAG([("A", "B"), ("B", "C")])
    dag.add_nodes_from(["A", "B", "C"], weights=[1, 2, 3])

    assert dag.get_parents("C") == ["B"]
    assert set(dag.get_children("A")) == {"B"}
    assert dag.get_leaves() == ["C"]

    modified = dag.do("B")
    assert not modified.has_edge("A", "B")


def test_active_trail_and_dconnection():
    dag = DAG([("A", "B"), ("B", "C")])
    trails = dag.active_trail_nodes("A")
    assert set(trails["A"]) == {"A", "B", "C"}

    blocked = dag.active_trail_nodes("A", observed=["B"])
    assert "C" not in blocked["A"]
    assert dag.is_dconnected("A", "C")
    assert not dag.is_dconnected("A", "C", observed=["B"])


def test_minimal_dseparator_and_ancestors():
    dag = DAG([("A", "C"), ("B", "C"), ("C", "D")])
    with pytest.raises(ValueError):
        dag.minimal_dseparator("A", "C")

    separator = dag.minimal_dseparator("A", "D")
    assert separator == {"C"}

    with pytest.raises(ValueError):
        dag._get_ancestors_of("X")


def test_get_random_and_roots():
    dag = DAG.get_random(n_nodes=3, edge_prob=0.0, latents=True)
    assert len(list(dag.nodes())) == 3
    assert set(dag.get_roots()) == set(dag.nodes())



import numpy as np
import pytest

from causalexplain.estimators.notears import linear


@pytest.mark.parametrize("loss_type", ["l2", "logistic", "poisson"])
def test_notears_linear_supported_losses(monkeypatch, loss_type):
    calls = {"minimize": 0}

    def fake_minimize(func, w_est, method=None, jac=None, bounds=None):
        calls["minimize"] += 1
        # Call func once to exercise _loss/_h branches
        func(w_est)
        result = type("Result", (), {})()
        result.x = w_est
        return result

    monkeypatch.setattr(linear.sopt, "minimize", fake_minimize)
    monkeypatch.setattr(linear.slin, "expm",
                        lambda mat: np.eye(mat.shape[0]) + mat)

    X = np.array([[0.1, 0.2], [0.3, 0.4]])
    W_est = linear.notears_linear(
        X, lambda1=0.0, loss_type=loss_type, max_iter=2, h_tol=1e-8)

    assert W_est.shape == (2, 2)
    assert np.allclose(W_est, np.zeros_like(W_est))
    assert calls["minimize"] >= 1


def test_notears_linear_unknown_loss(monkeypatch):
    def fake_minimize(func, w_est, method=None, jac=None, bounds=None):
        func(w_est)
        return type("Res", (), {"x": w_est})

    monkeypatch.setattr(linear.sopt, "minimize", fake_minimize)
    monkeypatch.setattr(linear.slin, "expm",
                        lambda mat: np.eye(mat.shape[0]) + mat)

    X = np.array([[0.1]])
    with pytest.raises(ValueError):
        linear.notears_linear(X, lambda1=0.0, loss_type="invalid")



import sys
import types

import numpy as np
import pandas as pd
import pytest

from causalexplain.independence import mic


class DummyProgBar:
    def start_subtask(self, *args, **kwargs):
        return types.SimpleNamespace(update_subtask=lambda *a, **k: None)


def test_pairwise_mic_raises_when_minepy_missing(monkeypatch):
    monkeypatch.setattr(mic, "ProgBar", lambda *args, **kwargs: DummyProgBar())
    monkeypatch.delitem(sys.modules, "minepy", raising=False)
    with pytest.raises(ModuleNotFoundError):
        mic.pairwise_mic(pd.DataFrame({"a": [0.1, 0.2], "b": [0.3, 0.4]}))


def test_pairwise_mic_with_stub(monkeypatch):
    data = pd.DataFrame({"a": [0.0, 1.0], "b": [1.0, 0.0], "c": [0.5, 0.5]})
    mic_values = np.array([0.1, 0.2, 0.3])
    tic_values = np.array([0.4, 0.5, 0.6])

    monkeypatch.setattr(mic, "ProgBar", lambda *args, **kwargs: DummyProgBar())
    fake_minepy = types.SimpleNamespace(pstats=lambda *a, **k: (mic_values, tic_values))
    monkeypatch.setitem(sys.modules, "minepy", fake_minepy)

    result = mic.pairwise_mic(data, prog_bar=False)
    assert result.shape == (3, 3)
    assert np.isclose(result[0, 1], mic_values[0])



from causalexplain.common import DEFAULT_REGRESSORS
import argparse
import json
import os
import pickle
import sys
from types import SimpleNamespace
from unittest.mock import MagicMock

import networkx as nx
import pandas as pd
import pytest

import causalexplain.causalexplainer as gd_module

sys.modules.setdefault("causalexplainer", gd_module)

from causalexplain import __main__ as main_mod  # noqa: E402
from causalexplain.causalexplainer import GraphDiscovery  # noqa: E402


@pytest.fixture
def sample_csv(tmp_path):
    df = pd.DataFrame({'X': [1, 2], 'Y': [3, 4], 'Z': [5, 6]})
    path = tmp_path / "sample.csv"
    df.to_csv(path, index=False)
    return str(path)


@pytest.fixture
def args_factory():
    def _factory(**overrides):
        base = argparse.Namespace(
            dataset=None,
            method='rex',
            true_dag=None,
            load_model=None,
            no_train=False,
            threshold=None,
            union=None,
            iterations=None,
            bootstrap=None,
            regressor=None,
            prior=None,
            seed=None,
            quiet=False,
            verbose=False,
            save_model=None,
            output=None
        )
        for key, value in overrides.items():
            setattr(base, key, value)
        return base
    return _factory


def test_parse_args_splits_comma_lists(monkeypatch):
    argv = [
        "prog",
        "-d", "data.csv",
        "-u", "a,b",
        "-r", "x,y",
    ]
    monkeypatch.setattr(sys, "argv", argv)
    args = main_mod.parse_args()
    assert args.union == ['a', 'b']
    assert args.regressor == ['x', 'y']
    assert args.dataset == "data.csv"


def test_check_args_requires_dataset_or_model(args_factory):
    args = args_factory()
    with pytest.raises(ValueError):
        main_mod.check_args_validity(args)


def test_check_args_with_dataset_and_save_defaults(
        sample_csv, args_factory, monkeypatch):
    monkeypatch.setattr(
        main_mod.utils, "valid_output_name",
        lambda filename, path: os.path.join(path, "unique_name.pickle"))
    args = args_factory(dataset=sample_csv, save_model='', seed=7)
    run_values = main_mod.check_args_validity(args)
    assert run_values['dataset_name'] == "sample"
    assert run_values['dataset_filepath'] == sample_csv
    assert run_values['regressors'] == DEFAULT_REGRESSORS
    assert run_values['seed'] == 7
    assert run_values['save_model'] == os.path.basename(
        sample_csv).replace('.csv', '') + "_rex.pickle"
    assert run_values['output_path'] == os.getcwd()
    assert run_values['bootstrap_iterations'] == main_mod.DEFAULT_BOOTSTRAP_TRIALS


def test_check_args_load_model_without_dataset(tmp_path, args_factory, monkeypatch):
    monkeypatch.chdir(tmp_path)
    model_path = tmp_path / "model.pkl"
    model_path.write_bytes(b"data")
    args = args_factory(load_model="model.pkl", no_train=True)
    run_values = main_mod.check_args_validity(args)
    assert run_values['dataset_name'] is None
    assert run_values['load_model'] == str(model_path)


def test_check_args_fails_when_load_model_missing(tmp_path, args_factory):
    args = args_factory(load_model=str(
        tmp_path / "missing.pkl"), no_train=True)
    with pytest.raises(FileNotFoundError):
        main_mod.check_args_validity(args)


def test_check_args_validates_method(sample_csv, args_factory):
    args = args_factory(dataset=sample_csv, method='invalid')
    with pytest.raises(AssertionError):
        main_mod.check_args_validity(args)


def test_check_args_handles_true_dag_and_prior(
        sample_csv, args_factory, tmp_path, monkeypatch):
    dot_file = tmp_path / "truth.dot"
    dot_file.write_text("digraph G { A -> B; }")
    prior_file = tmp_path / "prior.json"
    prior_data = [["A", "B"]]
    prior_file.write_text(json.dumps({"prior": prior_data}))
    monkeypatch.setattr(
        main_mod.utils, "graph_from_dot_file", lambda path: "graph")
    args = args_factory(
        dataset=sample_csv,
        method='pc',
        true_dag=str(dot_file),
        prior=str(prior_file))
    run_values = main_mod.check_args_validity(args)
    assert run_values['true_dag'] == str(dot_file)
    assert run_values['prior'] == prior_data
    assert run_values['ref_graph'] == "graph"
    assert run_values['regressors'] == ['pc']


def test_header_prints_banner(capsys):
    main_mod.header_()
    output = capsys.readouterr().out
    assert main_mod.HEADER_ASCII.strip() in output


def test_show_run_values_outputs_dataframe_shape(capsys):
    run_values = {'data': pd.DataFrame({'a': [1], 'b': [2]}), 'value': 5}
    main_mod.show_run_values(run_values)
    captured = capsys.readouterr().out
    assert "1x2 DataFrame" in captured
    assert "- value: 5" in captured


def test_main_trains_and_saves(monkeypatch, tmp_path):
    class DummyDiscovery:
        def __init__(self, **kwargs):
            self.init_args = kwargs
            self.trainer = {'initial': SimpleNamespace(
                dag="start", metrics=None)}
            self.saved = None
            self.printed = None

        def create_experiments(self):
            self.created = True

        def fit_experiments(self, *args):
            self.fit_args = args

        def combine_and_evaluate_dags(self, prior):
            self.combined_prior = prior
            return SimpleNamespace(dag="final_dag", metrics="final_metrics")

        def save(self, path):
            self.saved = path

        def printout_results(self, dag, metrics):
            self.printed = (dag, metrics)

    run_values = {
        'dataset_name': 'sample',
        'estimator': 'rex',
        'dataset_filepath': 'data.csv',
        'true_dag': None,
        'verbose': False,
        'seed': 7,
        'load_model': None,
        'no_train': False,
        'hpo_iterations': 3,
        'bootstrap_iterations': 4,
        'prior': [['A', 'B']],
        'output_path': str(tmp_path),
        'model_filename': str(tmp_path / "saved.pkl"),
        'output_dag_file': str(tmp_path / "dag.dot"),
    }
    times = iter([100.0, 101.0])
    monkeypatch.setattr(main_mod.time, "time", lambda: next(times))
    monkeypatch.setattr(main_mod.utils, "format_time",
                        lambda delta: (delta, "seconds"))
    saved_paths = []
    monkeypatch.setattr(main_mod.utils, "graph_to_dot_file",
                        lambda dag, path: saved_paths.append((dag, path)))
    instances = []

    def factory(**kwargs):
        inst = DummyDiscovery(**kwargs)
        instances.append(inst)
        return inst

    monkeypatch.setattr(main_mod, "GraphDiscovery", factory)
    monkeypatch.setattr(main_mod, "parse_args", lambda: SimpleNamespace())
    monkeypatch.setattr(main_mod, "check_args_validity", lambda _: run_values)
    main_mod.main()
    dummy = instances[0]
    assert dummy.saved == run_values['model_filename']
    assert saved_paths[0][1] == run_values['output_dag_file']


def test_main_loads_existing_model(monkeypatch):
    class DummyDiscovery:
        def __init__(self, **kwargs):
            self.trainer = {'one': SimpleNamespace(
                dag='loaded', metrics='metrics')}

        def load(self, path):
            self.loaded = path

        def fit_experiments(self, *args):
            pass

        def combine_and_evaluate_dags(self, prior):
            return SimpleNamespace(dag="combined", metrics="metrics")

        def printout_results(self, dag, metrics):
            self.printed = (dag, metrics)

        def save(self, path):
            self.saved = path

    run_values = {
        'dataset_name': 'sample',
        'estimator': 'rex',
        'dataset_filepath': None,
        'true_dag': None,
        'verbose': False,
        'seed': 0,
        'load_model': 'model.pkl',
        'no_train': True,
        'hpo_iterations': 0,
        'bootstrap_iterations': 0,
        'prior': None,
        'output_path': None,
        'model_filename': None,
        'output_dag_file': None,
    }
    instances = []

    def factory(**kwargs):
        inst = DummyDiscovery(**kwargs)
        instances.append(inst)
        return inst

    monkeypatch.setattr(main_mod, "GraphDiscovery", factory)
    monkeypatch.setattr(main_mod, "parse_args", lambda: SimpleNamespace())
    monkeypatch.setattr(main_mod, "check_args_validity", lambda _: run_values)
    monkeypatch.setattr(main_mod.time, "time", lambda: 0.0)
    monkeypatch.setattr(main_mod.utils, "format_time",
                        lambda delta: (delta, "seconds"))
    main_mod.main()
    assert instances[0].loaded == 'model.pkl'


def make_graph_discovery(sample_csv, tmp_path, model_type='rex'):
    return GraphDiscovery(
        experiment_name="exp",
        model_type=model_type,
        csv_filename=sample_csv,
        true_dag_filename=None,
        verbose=False,
        seed=1
    )


def test_fit_experiments_non_rex_calls_fit(sample_csv, tmp_path):
    gd = make_graph_discovery(sample_csv, tmp_path, model_type='pc')
    trainer = MagicMock()
    gd.trainer = {f"{gd.dataset_name}_pc": trainer}
    gd.fit_experiments(hpo_iterations=5, bootstrap_iterations=6, extra="value")
    trainer.fit_predict.assert_called_once_with(
        estimator='pc', verbose=False, extra="value")


def test_fit_experiments_rex_skips_rex_named_entries(sample_csv, tmp_path):
    gd = make_graph_discovery(sample_csv, tmp_path)
    rex_trainer = MagicMock()
    other_trainer = MagicMock()
    gd.trainer = {"exp_rex": rex_trainer, "exp_alt": other_trainer}
    gd.fit_experiments(hpo_iterations=2, bootstrap_iterations=3)
    other_trainer.fit_predict.assert_called_once()
    rex_trainer.fit_predict.assert_not_called()


def test_combine_and_evaluate_non_rex_sets_metrics(sample_csv, tmp_path, monkeypatch):
    gd = make_graph_discovery(sample_csv, tmp_path, model_type='pc')
    gd.ref_graph = nx.DiGraph()
    gd.data_columns = ['X', 'Y']
    trainer = SimpleNamespace(pc=SimpleNamespace(
        dag="dag"), dag=None, metrics=None)
    gd.trainer = {f"{gd.dataset_name}_pc": trainer}
    monkeypatch.setattr(
        "causalexplain.causalexplainer.evaluate_graph",
        lambda ref, dag, cols: {"sid": 0})
    result = gd.combine_and_evaluate_dags()
    assert result.dag == "dag"
    assert gd.metrics == {"sid": 0}


def test_combine_and_evaluate_rex_combines(sample_csv, tmp_path, monkeypatch):
    gd = make_graph_discovery(sample_csv, tmp_path)

    class Estimator:
        def __init__(self, label):
            self.dag = f"dag_{label}"
            self.shaps = SimpleNamespace(shap_discrepancies=f"disc_{label}")
    gd.trainer = {
        "exp_a": SimpleNamespace(rex=Estimator('a')),
        "exp_b": SimpleNamespace(rex=Estimator('b')),
    }
    monkeypatch.setattr(
        "causalexplain.causalexplainer.utils.combine_dags",
        lambda *args, **kwargs: (None, None, "combined", None))
    result = gd.combine_and_evaluate_dags(prior=[['A', 'B']])
    assert result.dag == "combined"
    assert gd.dag == "combined"


def test_run_invokes_sequence(sample_csv, tmp_path):
    gd = make_graph_discovery(sample_csv, tmp_path)
    gd.create_experiments = MagicMock()
    gd.fit_experiments = MagicMock()
    gd.combine_and_evaluate_dags = MagicMock()
    gd.run(5, 6, prior=[['A', 'B']], option=True)
    gd.create_experiments.assert_called_once()
    gd.fit_experiments.assert_called_once()
    gd.combine_and_evaluate_dags.assert_called_once_with(prior=[['A', 'B']])


def test_save_validates_state(sample_csv, tmp_path):
    gd = make_graph_discovery(sample_csv, tmp_path)
    gd.trainer = {}
    with pytest.raises(AssertionError):
        gd.save(str(tmp_path / "model.pkl"))


def test_save_writes_model(sample_csv, tmp_path, monkeypatch):
    gd = make_graph_discovery(sample_csv, tmp_path)
    gd.trainer = {"t": SimpleNamespace()}
    saved = {}
    monkeypatch.setattr(
        "causalexplain.causalexplainer.utils.save_experiment",
        lambda name, path, trainer, overwrite: saved.setdefault("path", os.path.join(path, name)))
    gd.save(str(tmp_path / "model.pkl"))
    assert saved["path"].endswith("model.pkl")


def test_load_sets_properties(tmp_path):
    gd = GraphDiscovery()
    trainer_data = {'a': SimpleNamespace(dag="dag", metrics="metrics")}
    model_path = tmp_path / "trainer.pkl"
    with open(model_path, 'wb') as handle:
        pickle.dump(trainer_data, handle)
    loaded = gd.load(str(model_path))
    assert gd.dag == "dag"
    assert loaded == trainer_data


def test_printout_results_handles_empty_graph(capsys):
    gd = GraphDiscovery()
    graph = nx.DiGraph()
    gd.printout_results(graph, None)
    assert "Empty graph" in capsys.readouterr().out


def test_printout_results_lists_edges(capsys):
    gd = GraphDiscovery()
    graph = nx.DiGraph()
    graph.add_edge("A", "B")
    graph.add_edge("B", "C")
    gd.printout_results(graph, "metrics")
    output = capsys.readouterr().out
    assert "A -> B" in output and "Graph Metrics" in output


def test_export_delegates_to_utils(sample_csv, tmp_path, monkeypatch):
    gd = make_graph_discovery(sample_csv, tmp_path, model_type='pc')
    gd.trainer = {'a': SimpleNamespace(dag="dag")}
    exported = {}
    monkeypatch.setattr(
        "causalexplain.causalexplainer.utils.graph_to_dot_file",
        lambda dag, path: exported.setdefault("path", path))
    result = gd.export("file.dot")
    assert result == "file.dot"
    assert exported["path"] == "file.dot"


def test_plot_calls_plot_module(sample_csv, tmp_path, monkeypatch):
    gd = make_graph_discovery(sample_csv, tmp_path, model_type='pc')
    gd.trainer = {'a': SimpleNamespace(dag="dag", ref_graph="ref")}
    called = {}
    monkeypatch.setattr(
        "causalexplain.causalexplainer.plot.dag",
        lambda **kwargs: called.setdefault("kwargs", kwargs))
    gd.plot(show_metrics=True, layout='circular')
    assert called["kwargs"]["graph"] == "dag"
    assert called["kwargs"]["layout"] == 'circular'


def test_model_property_returns_last_trainer(sample_csv, tmp_path):
    gd = make_graph_discovery(sample_csv, tmp_path, model_type='pc')
    gd.trainer = {'first': SimpleNamespace(
    ), 'second': SimpleNamespace(marker=True)}
    assert gd.model.marker is True



import numpy as np
import pytest

from causalexplain.estimators.notears import notears


def test_notears_standard_uses_custom_loss(monkeypatch):
    monkeypatch.setattr(notears.scipy.linalg, "expm",
                        lambda mat: np.eye(mat.shape[0]))

    calls = {"loss": 0, "grad": 0}

    def fake_minimize(func, w_flat, args=None, jac=None, method=None, options=None):
        class Res:
            def __init__(self, x):
                self.x = x
        return Res(w_flat)

    monkeypatch.setattr(notears.scipy.optimize, "minimize", fake_minimize)

    def custom_loss(W, data, cov, d, n):
        calls["loss"] += 1
        return 0.0

    def custom_grad(W, data, cov, d, n):
        calls["grad"] += 1
        return np.zeros_like(W)

    algo = notears.NOTEARS(name="test", loss=custom_loss, loss_grad=custom_grad)
    data = np.zeros((2, 2))
    out = algo.notears_standard(data, return_all_progress=False)

    assert calls["loss"] >= 1 and calls["grad"] >= 1
    assert out["W"].shape == (2, 2)
    assert out["h"] == 0.0


def test_fit_and_predict_flow(monkeypatch):
    model_output = {"W": np.array([[0.2, 0.0], [0.0, 0.0]])}

    def fake_notears_standard(self, data, return_all_progress=False):
        return model_output

    monkeypatch.setattr(notears.NOTEARS, "notears_standard", fake_notears_standard)
    monkeypatch.setattr(notears.utils, "graph_to_adjacency",
                        lambda ref_graph, labels: np.array([[0, 1], [0, 0]]))

    graph_inputs = {}
    monkeypatch.setattr(
        notears.utils,
        "graph_from_adjacency",
        lambda adjacency, node_labels: graph_inputs.update(
            {"adjacency": adjacency, "labels": node_labels}
        ) or "graph_obj")

    monkeypatch.setattr(notears, "evaluate_graph",
                        lambda ref_graph, dag: {"ref": ref_graph, "dag": dag})

    algo = notears.NOTEARS(name="pipeline")
    import pandas as pd

    train = pd.DataFrame([[0.0, 0.0]], columns=["a", "b"])
    ref_graph = "ref"
    algo.fit(train)
    algo.predict(ref_graph=ref_graph, threshold=0.1)

    assert np.array_equal(graph_inputs["adjacency"], np.array([[1, 0], [0, 0]]))
    assert graph_inputs["labels"] == ["a", "b"]
    assert algo.metrics == {"ref": ref_graph, "dag": "graph_obj"}
    assert algo.dag == "graph_obj"


def test_fit_predict_returns_graph(monkeypatch):
    monkeypatch.setattr(notears.NOTEARS, "fit",
                        lambda self, X: setattr(self, "model", {"W": np.zeros((1, 1))}) or self)
    monkeypatch.setattr(notears.NOTEARS, "predict",
                        lambda self, ref_graph=None, threshold=0.1: setattr(self, "dag", "dag_obj") or self)

    algo = notears.NOTEARS(name="fit-predict")
    dag = algo.fit_predict(train=np.array([[0.0]]), test=None, ref_graph=None)
    assert dag == "dag_obj"



"""
Here's an explanation of the changes:

1. We import `numpy` for array operations and `GradientBoostingRegressor` from 
scikit-learn as an equivalent to R's `glmboost`.
2. The function signature is similar, but we use `None` as the default for `pars` 
instead of an empty list.
3. We convert inputs to numpy arrays to ensure compatibility.
4. We center `y` by subtracting its mean.
5. We create and fit a `GradientBoostingRegressor`, which is similar to `glmboost` in R.
6. We create a dictionary `result` with the fitted values, residuals, and the model 
itself.
7. The `center=TRUE` parameter in the R version is not needed as scikit-learn's 
`GradientBoostingRegressor` handles feature centering internally.

Note that this Python version might not be exactly equivalent to the R version, 
as there could be differences in the underlying algorithms and default parameters.
You may need to adjust the `GradientBoostingRegressor` parameters to match the 
behavior of `glmboost` more closely if needed.
"""
import numpy as np
from sklearn.ensemble import GradientBoostingRegressor


def train_LMboost(X, y, pars=None):
    """_summary_

    Args:
        X (_type_): _description_
        y (_type_): _description_
        pars (dict, optional): Parameters

    Returns:
        dict: result
    """
    # Convert inputs to numpy arrays if they're not already
    X = np.array(X)
    y = np.array(y)

    # Center y
    y = y - np.mean(y)

    # Create and fit the model
    gb = GradientBoostingRegressor()
    gb.fit(X, y)

    # Prepare the result
    result = {
        'Yfit': gb.predict(X),
        'residuals': y - gb.predict(X),
        'model': gb
    }

    return result



"""
Application of early stage FCI rules to setup the skeleton of a causal graph to
an existing graph in order to prune potentially spurious edges.

(c) J. Renero, 2022
"""

import itertools
from collections import defaultdict

import networkx as nx
import numpy as np
import pygam
from hyppo.independence import Hsic
from pandas import DataFrame
from mlforge.progbar import ProgBar


# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid_name, C0116:missing-function-docstring, R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements, R1702:too-many-branches
# pylint: disable=W0106:expression-not-assigned


class GraphIndependence(object):
    """
    Class for removing independent edges from a causal graph.

    Parameters:
    -----------
    - base_graph (nx.DiGraph): The graph to be cleaned
    - condlen (int): The number of conditioning sets to use in the
        independence test. Defaults to 1, meaning that the cond. indep.
        test are of length=1.
    - condsize (int): The size of the conditioning sets to use in the
        independence test. Default is 0, which means that conditioning
        sets will be based on direct connections
    - prog_bar (bool): Whether to show a progress bar
    - verbose (bool): Whether to print debug messages
    - silent (bool): Whether to suppress all output

    Methods:
    --------
    - fit(X: DataFrame, y=None) -> Tuple[nx.DiGraph, Dict[str, List[Tuple[str, Tuple[str]]]]]:
        Remove edges from the graph that are conditionally independent.
    - predict() -> nx.DiGraph:
        Predicts the causal graph using the current independence tests and returns
        the resulting graph.
    - fit_predict(X: DataFrame, y=None) -> nx.DiGraph:
        Fits the model to the data and returns predictions.
    """

    def __init__(
            self,
            base_graph,
            condlen: int = 1,
            condsize: int = 0,
            max_rows: int = 500,
            prog_bar: bool = True,
            verbose: bool = False,
            silent: bool = False):

        self.base_graph = base_graph
        self.condlen = condlen
        self.condsize = condsize
        self.max_rows = max_rows
        self.prog_bar = prog_bar
        self.verbose = verbose
        self.silent_ = silent

        self._fit_desc = "Removing independent edges"

    def fit(self, X: DataFrame, y=None):
        """
        Remove edges from the graph that are conditionally independent.

        Arguments:
        ----------
        - base_graph (nx.DiGraph): The graph to be cleaned
        - data (DataFrame): The data used to discover the causal graph
        - condlen (int): The number of conditioning sets to use in the
            independence test. Defaults to 1, meaning that the cond. indep.
            test are of length=1.
        - condsize (int): The size of the conditioning sets to use in the
            independence test. Default is 0, which means that conditioning
            sets will be based on direct connections
        - prog_bar (bool): Whether to show a progress bar
        - verbose (bool): Whether to print debug messages

        Returns:
        --------
            nx.DiGraph: The cleaned graph
            actions: A dictionary of actions that were taken
        """
        # TODO: This was ON in the original code, but it seems to be a bug
        # condlen = 1
        # condsize = 0

        self.data = X.copy()
        self.y = y.copy() if y is not None else None

        # Check if base_graph contains data columns
        self.feature_names = list(self.data.columns)
        self.actions = defaultdict(list)
        self.sepset = {}
        for p in itertools.permutations(self.feature_names, 2):
            self.sepset[p] = ()
        # edge = list(self.base_graph.edges)

        self.G_skl = nx.DiGraph()
        self.G_skl.add_nodes_from(self.base_graph.nodes(data=True))
        self.G_skl.add_edges_from(self.base_graph.edges(data=True))

        # check if all column names in experiment.data correspond to nodes in G_skl.
        # If not add them
        for col in self.data.columns:
            if col not in self.G_skl.nodes():
                self.G_skl.add_node(col)

        pbar = ProgBar().start_subtask("Graph.Indep_fit", len(self.feature_names))
        for feature_idx, feature_name in enumerate(self.feature_names):
            self._remove_independent_edges(
                feature_name, self.condlen, self.condsize)
            pbar.update_subtask("Graph.Indep_fit", feature_idx+1)

        pbar.remove("Graph.Indep_fit")
        return self

    def predict(self):
        """
        Predicts the causal graph using the current independence tests and returns the resulting graph.

        Returns:
            The predicted causal graph.
        """
        return self.G_skl

    def fit_predict(self, X: DataFrame, y=None):
        """
        Fits the model to the data and returns predictions.

        Parameters:
        X (DataFrame): The input data to fit the model on.
        y (optional): The target variable to fit the model on.

        Returns:
        The predictions made by the model.
        """
        self.fit(X, y)
        return self.predict()

    def _gen_cond_sets(self, x, y, size):
        """ A function to build the set of conditioning sets to be for
        variables x and y on graph g of a certain size when generating a
        skeleton

        Arguments:
        ----------
        - x (str): One variable being tested for independence,
        - y (str): The other variable being tested for independence,
        - size (int): the size of each conditioning set to be returned

        Returns:
        --------
        A list of conditioning sets to be tested
        """
        # Handle size 0 case
        if size == 0:
            return [()]
        # Get all neighbors of x in g
        adjy = self.G_skl.neighbors(x)
        # Remove y from neighbour list
        adj = [node for node in adjy if node != y]
        # Generate all unique combinations of len size
        combos = list(itertools.combinations(adj, size))
        return combos

    @staticmethod
    def _residuals(x, y, Z):
        """
        Train a GAM with cond_set to predict "i", and then "j". For both cases
        get the residuals.

        Arguments:
        - x (np.array): x-values of the data
        - y (np.array): y-values of the data
        - Z (np.array): Z-values of the data

        Returns:
        - r_x (np.array): residuals of GAM fitted to predict x from Z
        - r_y (np.array): residuals of GAM fitted to predict y from Z
        """
        g = pygam.GAM()
        g.fit(Z, x)
        residuals_x = g.deviance_residuals(Z, x)
        g.fit(Z, y)
        residuals_y = g.deviance_residuals(Z, y)
        return residuals_x, residuals_y

    def _cond_indep_test(self, x: np.array, y: np.array, Z: np.array = None):
        """
        Perform the conditional independence test configured in the
        main class attribute

        Arguments:
        ----------
        - x (np.array): The first variable being tested for independence
        - y (np.array): The second variable being tested for independence
        - Z (np.array): optional (default=None) The conditioning set, if any

        Returns:
        --------
        - bool: True if x is independent of y, False otherwise
        - float: the t-statistic of the HSIC independence test
        - float: the p-value of the independence test
        """
        if x.shape[0] > self.max_rows:
            idx = np.random.choice(x.shape[0], self.max_rows, replace=False)
            x = x[idx]
            y = y[idx]
            if Z is not None:
                Z = Z[idx]

        if Z is None:
            stat, pval = Hsic().test(x, y)
        else:
            r_x, r_y = self._residuals(x, y, Z)
            stat, pval = Hsic().test(r_x, r_y)
        indep = pval > 0.05
        return indep, stat, pval

    def _update_graph(self, x, y, cs):
        print(
            f"    ! {x}⫫{y}|({','.join(cs)}) ❌ Independent!") if self.verbose else None
        self.G_skl.remove_edge(x, y)
        self.sepset[(x, y)] = cs
        self.sepset[(y, x)] = cs

    def _update_actions(self, tup):
        if tup is not None:
            if tup[2]:
                self.actions[tup[0]].append((tup[1], tup[2]))
            else:
                self.actions[tup[0]].append((tup[1], ()))

    @staticmethod
    def _empty(x: list):
        if len(x) == 0:
            return True
        else:
            return len(x[0]) == 0

    def _test_cond_independence(self, x_name, y_name, cond_sets):
        """
        Test if X and Y are independent, using all possible conditioning sets
        in the graph from nodes adjacent to X. If it is found that X and Y are
        independent, exploration finish, and the edge between X and Y is
        removed from the graph.
        """
        L = len(cond_sets)
        x = self.data.loc[:, x_name].values.reshape(-1, 1)
        y = self.data.loc[:, y_name].values.reshape(-1, 1)

        for idx, cond_set in enumerate(cond_sets):
            Z = self.data.loc[:, cond_set].values if not self._empty(
                cond_set) else None
            independent, _, _ = self._cond_indep_test(x, y, Z)
            if independent:
                self._update_graph(x_name, y_name, cond_set)
                return x_name, y_name, cond_set

            # Debug msg when not independent
            msg = "{}- [{:02d}/{:02d}] {}⫫{}|({}) - Not independent"
            msg = msg.format("    ", idx + 1, L, x_name,
                             y_name, ','.join(cond_set))
            print(msg) if self.verbose else None

        return None

    def _remove_independent_edges(self, x, condlen, condsize):
        if self.actions is None:
            self.actions = defaultdict(list)

        neighbors = list(self.G_skl.neighbors(x))
        neighbors_str = ",".join(list(neighbors))
        print(
            f"> Iterating over neighbors: {neighbors_str}") if self.verbose else None
        for ny, y in enumerate(neighbors):
            print(
                f"+ y = {y}; {ny + 1}/{len(neighbors)}") if self.verbose else None
            # Generate the conditioning sets needed for independence tests
            condSets = self._gen_cond_sets(x, y, condsize)
            if len(condSets) == 0:
                continue
            condlen += len(condSets)
            # Test for independence with each conditioning set
            cs = ",".join([f"({','.join(cs)})" for cs in condSets])
            msg = f"  > Check {x}⫫{y} over {len(condSets)} CS: [{cs}]"
            print(msg) if self.verbose else None
            tup = self._test_cond_independence(x, y, condSets)
            if tup is not None:
                self._update_actions(tup)

        return condlen

    def compute_cond_indep_pvals(self):
        """
        Perform the `_test_cond_independence` on all pairs of nodes in the graph, and
        store the p_value (3rd element in the return tuple) on the class
        `cond_indep_pvals`
        """
        self.cond_indep_pvals = {}
        if self.verbose and self.data.shape[0] > self.max_rows:
            print(
                f"[INFO] Data has more than {self.max_rows} rows. "
                f"Sampling {self.max_rows} rows for cond. indep. test.")

        for x in self.feature_names:
            for y in self.feature_names:
                if x == y:
                    continue
                X = self.data.loc[:, x].values.reshape(-1, 1)
                Y = self.data.loc[:, y].values.reshape(-1, 1)
                _, _, pval = self._cond_indep_test(X, Y, None)
                self.cond_indep_pvals[(x, y)] = pval

        return self.cond_indep_pvals



import networkx as nx
import numpy as np
import pandas as pd
import pytest

from causalexplain.explainability import hierarchies as hmod


def test_compute_correlation_matrix_branches(monkeypatch):
    data = pd.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
    corr = hmod.Hierarchies.compute_correlation_matrix(data, method="spearman")
    assert corr.loc["a", "b"] == pytest.approx(1.0)

    called = {}

    def fake_mic(df, alpha, c, prog_bar):
        called["args"] = (alpha, c, prog_bar)
        return pd.DataFrame(np.eye(df.shape[1]), columns=df.columns, index=df.columns)

    monkeypatch.setattr(hmod, "pairwise_mic", fake_mic)
    corr = hmod.Hierarchies.compute_correlation_matrix(data, method="mic", mic_alpha=0.7, mic_c=3, prog_bar=True)
    assert called["args"] == (0.7, 3, True)
    assert corr.equals(pd.DataFrame(np.eye(2), columns=["a", "b"], index=["a", "b"]))

    with pytest.raises(ValueError):
        hmod.Hierarchies.compute_correlation_matrix(data, method="unknown")


def test_compute_correlated_features_and_are_connected():
    corr = pd.DataFrame([[1.0, 0.9], [0.9, 1.0]], columns=["x", "y"], index=["x", "y"])
    correlated = hmod.Hierarchies.compute_correlated_features(corr, 0.8, ["x", "y"])
    assert correlated["x"] == ["y"]

    linkage_mat = np.array([[0, 1, 0.25, 2]])
    clusters = hmod.Hierarchies()._clusters_from_linkage(linkage_mat, ["x", "y"])
    assert hmod.Hierarchies()._are_connected(clusters, "x", "y") == pytest.approx(0.25)


def test_get_direction_and_directed_pair():
    g = nx.DiGraph()
    g.add_edge("a", "b")
    assert hmod._get_direction(g, "a", "b") == ("-->", 1)
    g.add_edge("b", "a")
    assert hmod._get_direction(g, "a", "b") == ("<->", 0)
    g.remove_edge("a", "b")
    assert hmod._get_direction(g, "a", "b") == ("<--", -1)
    with pytest.raises(ValueError):
        hmod._get_direction(g, "missing", "b")

    assert hmod._get_directed_pair(g, "a", "b") == ("b", "a")
    g.remove_edge("b", "a")
    assert hmod._get_directed_pair(g, "a", "b") is None


def test_connect_isolated_nodes_adds_edge():
    g = nx.DiGraph()
    # self-loops keep nodes present without defining direction between u and v
    g.add_edges_from([("u", "u"), ("v", "v")])
    linkage_mat = np.array([[0, 1, 0.1, 2]])
    connected = hmod.connect_isolated_nodes(g, linkage_mat, ["u", "v"], verbose=True)
    assert connected.has_edge("u", "v") or connected.has_edge("v", "u")


def test_connect_hierarchies_no_crash_with_clusters():
    g = nx.DiGraph()
    g.add_edge("a", "b")
    feature_names = ["a", "b", "c"]
    linkage_mat = np.array([[0, 1, 0.1, 2], [3, 2, 0.2, 2]])

    # The helper raises if nodes are missing; patch to only report orientation
    def safe_direction(graph, u, v):
        if not graph.has_node(u):
            graph.add_node(u)
        if not graph.has_node(v):
            graph.add_node(v)
        return " · ", None

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(hmod, "_get_direction", safe_direction)
    result = hmod.connect_hierarchies(g, linkage_mat, feature_names, verbose=True)
    monkeypatch.undo()
    assert isinstance(result, nx.DiGraph)



import torch

from causalexplain.models._loss import MMDLoss, RBF


def test_rbf_bandwidth_and_kernel_values():
    rbf = RBF()
    X = torch.tensor([[0.0], [1.0]])
    L2 = torch.cdist(X, X) ** 2

    # Auto-computed bandwidth from pairwise distances.
    assert rbf.get_bandwidth(L2) == 1.0

    # Fixed bandwidth should bypass the automatic path.
    fixed = RBF(bandwidth=0.5)
    assert fixed.get_bandwidth(L2) == 0.5

    kernel_matrix = rbf(X)
    assert kernel_matrix.shape == (2, 2)
    assert torch.all(kernel_matrix > 0)


def test_mmd_loss_balanced_samples_zero_gap():
    X = torch.tensor([[0.0], [1.0]])
    Y = torch.tensor([[0.0], [1.0]])

    loss_fn = MMDLoss()
    loss = loss_fn(X, Y)
    assert torch.isclose(loss, torch.tensor(0.0), atol=1e-6)



import numpy as np
import torch

from causalexplain.estimators.notears import locally_connected


def test_forward_without_bias_matches_numpy():
    torch.set_default_dtype(torch.double)
    n, d, m1, m2 = 2, 3, 2, 2
    input_array = np.arange(n * d * m1).reshape(n, d, m1).astype(float)
    weight = np.ones((d, m1, m2))

    layer = locally_connected.LocallyConnected(d, m1, m2, bias=False)
    with torch.no_grad():
        layer.weight[:] = torch.from_numpy(weight)

    output = layer(torch.from_numpy(input_array)).detach().numpy()
    expected = np.zeros((n, d, m2))
    for j in range(d):
        expected[:, j, :] = input_array[:, j, :] @ weight[j]

    assert np.allclose(output, expected)


def test_forward_with_bias_adds_offset():
    torch.set_default_dtype(torch.double)
    layer = locally_connected.LocallyConnected(1, 2, 2, bias=True)
    with torch.no_grad():
        layer.weight[:] = torch.tensor([[[1.0, -1.0], [0.5, 0.5]]])
        layer.bias[:] = torch.tensor([[0.2, 0.3]])

    inp = torch.tensor([[[2.0, 1.0]]])  # shape [1, d=1, m1=2]
    output = layer(inp).detach().numpy()

    manual = np.array([[[2.0 * 1.0 + 1.0 * 0.5 + 0.2,
                         2.0 * -1.0 + 1.0 * 0.5 + 0.3]]])
    assert np.allclose(output, manual)



import pandas as pd
import pytest

from causalexplain.estimators.pc import estimators


def test_convert_args_tuple_wraps_parents():
    calls = {}

    @estimators.convert_args_tuple
    def dummy(_self, variable, parents, complete_samples_only, weighted):
        calls["parents"] = parents
        return parents

    result = dummy(object(), "v", ["a", "b"], None, False)
    assert isinstance(result, tuple)
    assert calls["parents"] == ("a", "b")


def test_init_data_collects_state_names_and_validates():
    data = pd.DataFrame({"A": ["x", "x", "y"], "B": ["a", "b", "a"]})
    est = estimators.StructureEstimator()
    est._init_data(data=data, state_names=None, complete_samples_only=False)
    assert est.state_names["A"] == ["x", "y"]

    with pytest.raises(ValueError):
        est._init_data(data=data, state_names={"A": ["x"], "B": ["a", "b"]})


def test_state_counts_with_and_without_parents():
    data = pd.DataFrame({"A": ["x", "x", "y", "y"], "B": ["a", "b", "a", "b"]})
    est = estimators.StructureEstimator()
    est._init_data(data=data, complete_samples_only=True)

    counts = est.state_counts("A")
    assert counts.loc["x", "A"] == 2
    assert counts.loc["y", "A"] == 2

    counts_with_parent = est.state_counts("B", parents=["A"])
    assert counts_with_parent.loc["a", ("x",)] == 1
    assert counts_with_parent.loc["b", ("y",)] == 1


def test_state_counts_handles_missing_and_weighted_flag():
    data = pd.DataFrame({"A": ["x", None], "B": ["a", "b"]})
    est = estimators.StructureEstimator()
    est._init_data(data=data, complete_samples_only=False)

    counts = est.state_counts("B", parents=["A"], complete_samples_only=False)
    assert counts.loc["a", ("x",)] == 1

    with pytest.raises(ValueError):
        est.state_counts("A", weighted=True)



import numpy as np

from causalexplain.estimators.cam import selLasso as sel_lasso_module


def test_sel_lasso_marks_nonzero_coefficients(monkeypatch):
    monkeypatch.setattr(
        sel_lasso_module,
        "train_lasso",
        lambda X, y, pars=None: {"model": type("M", (), {"coef_": np.array([1.0, 0.0])})()},
    )

    X = np.array([[1.0, 2.0, 3.0], [1.5, 2.5, 3.5]])
    selection = sel_lasso_module.selLasso(X, k=1)

    assert selection == [True, False, False]


def test_sel_lasso_single_variable():
    X = np.ones((2, 1))
    assert sel_lasso_module.selLasso(X, k=0) == []



import numpy as np

from causalexplain.estimators.cam import selGam as sel_gam_module


def test_sel_gam_marks_significant_predictors(monkeypatch):
    monkeypatch.setattr(
        sel_gam_module,
        "train_gam",
        lambda X, y, pars=None, verbose=False: {"p_values": np.array([0.0001, 0.5, 0.9])},
    )

    X = np.array([[1.0, 2.0, 3.0], [1.5, 2.5, 3.5]])
    selection = sel_gam_module.selGam(X, pars={"cutOffPVal": 0.001, "numBasisFcts": 5}, k=2)

    assert bool(selection[0]) is True
    assert bool(selection[1]) is False
    assert len(selection) == 3


def test_sel_gam_no_selection_for_single_feature():
    X = np.array([[1.0], [2.0]])
    assert sel_gam_module.selGam(X) == []



"""
Key differences and notes:

- We use NumPy arrays instead of R matrices.
- The pygam library is used instead of R's gam function.
- The formula creation is different. In pygam, we create a list of smooth terms.
- Error handling is done with a try-except block instead of R's try().
- The df, edf, and edf1 calculations are approximations, as pygam doesn't provide
    exact equivalents to R's GAM implementation.
- The function signature includes type hints for better code clarity.

To use this function, you'll need to install the required libraries: pygam

This Python version should provide similar functionality to the R version,
but there might be some differences in the exact numerical results due to the
different implementations of GAM in R and Python.
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, W0221:arguments-differ
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, E0401:import-error
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from typing import Any, Dict

import numpy as np
from pygam import GAM, terms
from statsmodels.gam.api import BSplines, GLMGam


def _build_gam_formula(p, num_splines, lam=False):
    """
    Create a pyGAM formula for a GAM model with 'p' smooth terms, each with
    'num_splines' basis functions. If 'lam' is True, the smooth terms are
    also regularized with a penalty term of 0, which is different from
    R's `gam` function.

    Parameters
    ----------
    p : int
        The number of smooth terms in the GAM model.
    num_splines : int
        The number of basis functions for each smooth term.
    lam : bool, optional
        Whether to add a penalty term to each smooth term. Defaults to
        False.

    Returns
    -------
    formula : pyGAM formula
        The pyGAM formula for the GAM model.
    """
    gam_formula = None
    for i in range(p):
        if gam_formula is None:
            if not lam:
                gam_formula = terms.s(i, n_splines=num_splines)
            else:
                gam_formula = terms.s(i, n_splines=num_splines, lam=0)
        else:
            if not lam:
                gam_formula = gam_formula + \
                    terms.s(i, n_splines=num_splines)
            else:
                gam_formula = gam_formula + \
                    terms.s(i, n_splines=num_splines, lam=0)

    return gam_formula


def train_gam(
        X: np.ndarray,
        y: np.ndarray,
        pars: Dict[str, Any] = None,
        verbose: bool=False) -> Dict[str, Any]:
    """
    Train a Generalized Additive Model using pyGAM.

    Parameters
    ----------
    X : np.ndarray
        Input features.
    y : np.ndarray
        Target variable.
    pars : Dict[str, Any], optional
        Model parameters. Defaults to None.

    Returns
    -------
    result : Dict[str, Any]
        Model results.
    """
    if pars is None:
        pars = {"num_basis_fcts": 10}
    if "num_basis_fcts" not in pars:
        pars["num_basis_fcts"] = 10

    X = np.asarray(X)
    y = np.asarray(y).reshape(-1, 1)

    if verbose:
        print(f". . . . Fitting GAM on {X.shape[1]} variables")
        print(". . . . . First 8 elements of each column of X:")
        for col in range(X.shape[1]):
            print(f". . . . . . Column {col}: {X[:8, col]}")
        print(f". . . . . First 8 elements of Y: {y[:8].flatten()}")

    n, p = X.shape

    if n / p < 3 * pars["num_basis_fcts"]:
        pars["num_basis_fcts"] = int(np.ceil(n / (3 * p)))
        print(
            f"Changed number of basis functions to {pars['num_basis_fcts']} in order "
            f"to have enough samples per basis function") if verbose else None
    gam_formula = _build_gam_formula(p, pars["num_basis_fcts"])
    try:
        expr = ""
        for i in range(p):
            expr += f"s({i})"
            if i == p-1:
                break
            expr += " + "
        print(f". . . . . Fitting GAM: {expr}") if verbose else None
        model = GAM(gam_formula)
        model.fit(X, y)
    except Exception:
        print(". . . . . There was some error with GAM. The smoothing parameter is set to zero.")
        gam_formula = _build_gam_formula(p, pars["num_basis_fcts"], lam=True)
        model = GAM(gam_formula)
        model.fit(X, y)

    result = {
        "Yfit": model.predict(X).reshape(-1, 1),
        "residuals": (y - model.predict(X)).reshape(-1, 1),
        "model": model,
        # Approximate
        "df": model.statistics_["n_samples"] - model.statistics_["edof"],
        "edf": model.statistics_["edof"],
        # pygam doesn't have a direct equivalent to edf1
        "edf1": model.statistics_["edof"],
        "p_values": model.statistics_["p_values"]
    }

    return result


# ... existing code ...


def train_gam_sm(
        X: np.ndarray,
        y: np.ndarray,
        pars: Dict[str, Any] = None) -> Dict[str, Any]:
    """Train a Generalized Additive Model using statsmodels.gam.

    Args:
        X (np.ndarray): Input features.
        y (np.ndarray): Target variable.
        pars (Dict[str, Any], optional): Model parameters. Defaults to None.

    Returns:
        Dict[str, Any]: Model results.
    """

    if pars is None:
        pars = {"num_basis_fcts": 10}
    if "num_basis_fcts" not in pars:
        pars["num_basis_fcts"] = 10

    y = y.reshape(-1, 1)  # Change this line to flatten y

    print(f". . . . Fitting GAM on {X.shape[1]} variables")
    print(". . . . . First 8 elements of each column of X:")
    for col in range(X.shape[1]):
        print(f". . . . . . Column {col}: {X[:8, col]}")
    print(f". . . . . First 8 elements of Y: {y[:8].flatten()}")

    n, p = X.shape

    if n / p < 3 * pars["num_basis_fcts"]:
        pars["num_basis_fcts"] = int(np.ceil(n / (3 * p)))
        print(
            f"Changed number of basis functions to {pars['num_basis_fcts']} in order "
            f"to have enough samples per basis function")

    # Create B-spline basis
    basis = BSplines(X, df=pars["num_basis_fcts"] * p, degree=[3] * p)

    try:
        mod_gam = GLMGam(y, smoother=basis)
        mod_gam.fit()
    except Exception:
        print(". . . . . There was some error with GAM. The smoothing parameter is set to zero.")
        # Adjust basis for zero smoothing (if applicable)
        basis = BSplines(X, df=[1] * p, degree=[3] * p)
        mod_gam = GLMGam(y, smoother=basis)
        mod_gam.fit()

    result = {
        # Ensure Yfit is reshaped correctly
        "Yfit": mod_gam.predict(X),
        # Ensure residuals are reshaped correctly
        "residuals": (y - mod_gam.predict(X)).reshape(-1, 1),
        "model": mod_gam,
        # Approximate
        "df": mod_gam.df_resid,
        "edf": mod_gam.df_model,
        # pygam doesn't have a direct equivalent to edf1
        "edf1": mod_gam.df_model
    }

    return result



import numpy as np

from causalexplain.estimators.cam import selLm as sel_lm_module


def test_sel_lm_uses_p_values(monkeypatch):
    class FakeResult:
        pvalue = np.array([0.5, 0.0001, 0.7])

    monkeypatch.setattr(sel_lm_module.stats, "linregress", lambda X, y: FakeResult())

    X = np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]])
    selection = sel_lm_module.selLm(X, pars={"cut_off_p_val": 0.001}, k=1)

    assert selection == [True, False, False]


def test_sel_lm_single_variable():
    X = np.ones((2, 1))
    assert sel_lm_module.selLm(X, k=0) == []



import numpy as np
import torch
import scipy.linalg as slin

from causalexplain.estimators.notears.trace_expm import trace_expm


def test_trace_expm_forward_and_backward():
    torch.set_default_dtype(torch.double)
    matrix = torch.tensor([[1.0, 0.0], [0.0, 1.0]], requires_grad=True)
    result = trace_expm(matrix)
    expected = slin.expm(matrix.detach().numpy()).trace()
    assert np.isclose(result.item(), expected)

    scalar = 0.5 * result * result
    scalar.backward()
    expected_grad = result.item() * slin.expm(matrix.detach().numpy()).T
    assert np.allclose(matrix.grad.detach().numpy(), expected_grad)



"""
- R's `cat` is replaced with Python's `print`.
- R's `stop` is replaced with Python's `raise` for exceptions.
- R's `!` is replaced with Python's `~` for logical negation.
- R's `prod` is replaced with `np.prod` from NumPy.
- R's `var` is replaced with `np.var` from NumPy.
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, W0221:arguments-differ
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, E0401:import-error
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches


import numpy as np
from causalexplain.estimators.cam.train_gam import train_gam
from causalexplain.estimators.cam.train_linear import train_linear


def computeScoreMatParallel(
        row_parents,
        score_name,
        X,
        sel_mat,
        verbose,
        node2,
        i,
        pars_score,
        interv_mat,
        interv_data):

    # the i-th row of row_parents contains possible parents of node2 (we call
    # them "parents_to_check")
    parents_to_check = row_parents[i, :]
    if verbose:
        print(f"parents_to_check: {parents_to_check.flatten()}")

    if verbose:
        print(
            f". Compute score entry for regressing {node2} on {parents_to_check}")

    if interv_data:
        X2 = X[~interv_mat[:, node2], :]
    else:
        X2 = X

    parents_to_check = parents_to_check.astype(int)
    if node2 not in parents_to_check and \
            np.prod(sel_mat[parents_to_check, node2]) == 1:
        if verbose:
            print(". . node2 not in parentsToCheck AND all parents are selected")
            print(
                f". . . '{node2}' not in parents_to_check ({parents_to_check}) OR\n"
                f". . . selMat[parentsToCheck, node2] = "
                f"{sel_mat[parents_to_check, node2]}")
        if score_name == "SEMSEV":
            raise ValueError("This score does not work. It does not decouple.")
        elif score_name == "SEMIND":
            raise NotImplementedError("NOT IMPLEMENTED YET")
        elif score_name == "SEMGAM":
            if verbose:
                print(f". . . parentsToCheck: {parents_to_check.flatten()}\n"
                    f". . . node2: {node2}")
            mod_gam = train_gam(X2[:, parents_to_check],
                                X2[:, node2], pars=pars_score, verbose=verbose)
            score = -np.log(np.var(mod_gam['residuals']))
            print(f". . . score: {score}") if verbose else None
        elif score_name == "SEMLIN":
            mod_gam = train_linear(X2[:, parents_to_check], X2[:, node2])
            score = -np.log(np.var(mod_gam['residuals']))
        else:
            raise ValueError("I do not know this score function.")
    else:
        print(". . NOT matching\n"
              f". . . '{node2}' not in? parents_to_check ({parents_to_check}) "
              f"{node2 not in parents_to_check} OR\n"
              f". . . selMat[parentsToCheck, node2] = "
              f"{sel_mat[parents_to_check, node2]}") if verbose else None
        score = -np.inf

    return score



import numpy as np

from causalexplain.estimators.cam import train_GAMboost


def test_train_gamboost_builds_spline_and_model():
    X = np.linspace(0, 1, 8).reshape(-1, 1)
    y = np.sin(2 * np.pi * X).ravel()

    result = train_GAMboost.train_GAMboost(X, y)

    assert set(result) == {"Yfit", "residuals", "model"}
    assert result["Yfit"].shape == y.shape
    assert result["residuals"].shape == y.shape



import numpy as np
import pytest

from causalexplain.estimators.ges import utils


def test_neighbors_adj_pa_ch():
    A = np.array([
        [0, 1, 1],
        [1, 0, 0],
        [0, 1, 0],
    ])
    assert utils.neighbors(0, A) == {1}
    assert utils.adj(0, A) == {1, 2}
    assert utils.pa(2, A) == {0, 1}
    assert utils.ch(1, A) == {0, 2}


def test_na_filters_neighbors_adjacent_to_node():
    A = np.array([
        [0, 1],
        [1, 0],
    ])
    assert utils.na(0, 1, A) == {1}
    assert utils.na(1, 0, A) == {0}


def test_is_clique_and_skeleton():
    A = np.array([
        [0, 1, 1],
        [0, 0, 1],
        [0, 0, 0],
    ])
    assert utils.is_clique({0, 1, 2}, A) is False
    assert utils.is_clique({0, 1}, A) is True


def test_is_dag_and_topological_ordering():
    dag = np.array([
        [0, 1, 0],
        [0, 0, 1],
        [0, 0, 0],
    ])
    assert utils.is_dag(dag)
    order = utils.topological_ordering(dag)
    assert order == [0, 1, 2]

    cyclic = np.array([
        [0, 1],
        [1, 0],
    ])
    assert utils.is_dag(cyclic) is False
    with pytest.raises(ValueError):
        utils.topological_ordering(cyclic)



"""
Hierarchy of links

Can I use the information above to decide wether to connect groups
of variables linked together?

"""
# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, E1121:too-many-function-args
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from collections import defaultdict
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd

from scipy.cluster.hierarchy import dendrogram, fcluster, linkage
from scipy.spatial.distance import squareform

from causalexplain.independence.mic import pairwise_mic


class Hierarchies:
    """
    Class representing the hierarchy of links between variables.

    Parameters
    ----------
    method : str or Callable, optional
        Method to use to compute the correlation. Default is 'spearman', but can
        also be 'pearson', 'kendall' or 'mic'.
    alpha : float, optional
        Threshold for the correlation. Default is 0.6.
    c : int, optional
        Number of clusters to be formed. Default is 15. Only valid with MIC.
    linkage_method : str, optional
        Method to use to compute the linkage. Default is 'complete'.
    correlation_th : float, optional
        Threshold for the correlation. Default is None.
    prog_bar : bool, optional
        Whether to show a progress bar during computation. Default is False.
    verbose : bool, optional
        Whether to print additional information. Default is False.
    silent : bool, optional
        Whether to suppress all output. Default is False.
    """

    correlations = None
    linkage_mat = None

    def __init__(
            self,
            method: str = 'spearman',
            mic_alpha: float = 0.6,
            mic_c: int = 15,
            linkage_method: str = 'complete',
            correlation_th: float = None,
            prog_bar: bool = False,
            verbose: bool = False,
            silent: bool = False):
        """
        Initialize the Hierarchies object.

        Parameters
        ----------
        method : str or Callable, optional
            Method to use to compute the correlation. Default is 'spearman',
            but can also be 'pearson', 'kendall' or 'mic'.
        mic_alpha : float, optional
            Threshold for the correlation. Default is 0.6.
        mic_c : int, optional
            Number of clusters to be formed. Default is 15. Only valid with MIC.
        linkage_method : str, optional
            Method to use to compute the linkage. Default is 'complete'.
        correlation_th : float, optional
            Threshold for the correlation. Default is None.
        prog_bar : bool, optional
            Whether to show a progress bar during computation. Default is False.
        verbose : bool, optional
            Whether to print additional information. Default is False.
        silent : bool, optional
            Whether to suppress all output. Default is False.
        """
        self.method = method
        self.alpha = mic_alpha
        self.c = mic_c
        self.linkage_method = linkage_method
        self.correlation_th = correlation_th
        self.prog_bar = prog_bar
        self.verbose_ = verbose
        self.silent_ = silent
        self.linkage_mat: np.ndarray = None

    def fit(self, X: pd.DataFrame):
        """
        Compute the hierarchy of links between variables using the correlation
        method specified in `corr_method`.

        Parameters
        ----------
        X : pd.DataFrame
            The input data.
        y : None
            Ignored.

        Returns
        -------
        self : Hierarchies
            The fitted Hierarchies object.
        """
        # copy X into self.data
        self.data = X.copy()
        self.feature_names = list(self.data.columns)

        # Set the list of correlated features for each target
        self.correlations = self.compute_correlation_matrix(
            self.data, method=self.method, mic_alpha=self.alpha, mic_c=self.c,
            prog_bar=self.prog_bar)
        self.correlated_features = self.compute_correlated_features(
            self.correlations, self.correlation_th, self.feature_names)

        # Compute the dissimilarity matrix
        self.dissimilarity = 1 - np.abs(self.correlations)
        close_to_zero = self.dissimilarity < 1.0e-6
        self.dissimilarity[close_to_zero] = 0.0
        self.linkage_mat = linkage(squareform(
            self.dissimilarity), self.linkage_method)

        return self

    @staticmethod
    def compute_correlation_matrix(
            data: pd.DataFrame,
            method='spearman',
            mic_alpha: float = 0.6,
            mic_c: int = 15,
            prog_bar=False):
        """
        Compute the correlation matrix.

        Parameters
        ----------
        data : pd.DataFrame
            The input data.
        method : str or Callable, optional
            Method to use to compute the correlation. Default is 'spearman',
            but can also be 'pearson', 'kendall' or 'mic'.
        prog_bar : bool, optional
            Whether to show a progress bar during computation. Default is False.

        Returns
        -------
        correlations : pd.DataFrame
            The correlation matrix.
        """
        if method in ['spearman', 'pearson', 'kendall']:
            correlations = data.corr(method=method)
        elif method == 'mic':
            correlations = pairwise_mic(
                data, alpha=mic_alpha, c=mic_c, prog_bar=prog_bar)
        else:
            raise ValueError(
                f"Unknown correlation method: {method}. Use 'spearman', "
                f"'pearson', 'kendall' or 'mic'.")

        return correlations

    @staticmethod
    def compute_correlated_features(correlations, correlation_th, feature_names, verbose=False):
        """
        Compute the list of correlated features for each target.

        Parameters
        ----------
        correlations : pd.DataFrame
            The correlation matrix.
        correlation_th : float
            Threshold for the correlation.
        feature_names : List[str]
            The list of feature names.
        verbose : bool, optional
            Whether to print additional information. Default is False.

        Returns
        -------
        correlated_features : defaultdict(list)
            The list of correlated features for each target.
        """
        correlated_features = defaultdict(list)
        if correlation_th:
            for target_name in feature_names:
                corr_features = list(
                    correlations[(correlations[target_name] > correlation_th)
                                 & (correlations[target_name] < 1.0)].index)
                if len(corr_features) > 0:
                    correlated_features[target_name] = corr_features
                    if verbose:
                        print(
                            f"CORRELATED FEATS for {target_name}: {corr_features}")

        return correlated_features

    def expand_clusters_perm_importance(self, pi, ground_truth=None):
        """
        Expand the clusters of the linkage matrix to include the features that are
        in the same cluster in the permutation importance matrix.
        It expands, for each cluster, with the metrics related to correlation,
        deltas, backward PI, etc. Used to determine if some criteria can be extracted.

        Parameters
        ----------
        pi : pd.DataFrame
            Permutation importance matrix.
        ground_truth : pd.DataFrame, optional
            Ground truth matrix.

        Returns
        -------
        None
        """
        clusters = self._clusters_from_linkage(
            self.linkage_mat, self.data.columns)
        correlations = np.abs(self.data.corr("spearman"))
        for i, feature in enumerate(self.data.columns):
            print(f"{feature}")
            conn = "└─"
            for j, (name, fwd_PI) in enumerate(pi.feature_importances_[feature].items()):
                # check if this the last item in the list
                conn = "└─> " if j == len(pi.feature_importances_[
                    feature]) - 1 else "├─> "
                print(f" {conn}{name:>4s}: {fwd_PI:.02f}", end="")
                if feature in pi.feature_importances_[name].keys():
                    bwd_PI = pi.feature_importances_[name][feature]
                    print(
                        f" r({bwd_PI:.02f},∂={np.abs(fwd_PI-bwd_PI):.02f})", end="")
                    reverse = True
                else:
                    print(" r(0.00,∂=0.00)", end="")
                    reverse = False
                print(f" | c({correlations.loc[feature, name]:.02f})", end="")
                print(f" | R2({pi.regression_importances_[i]:.2f})", end="")
                degree = self._are_connected(clusters, feature, name)
                connected = True if degree is not None else False
                if connected:
                    print(f" | d{degree:.02f}", end="")
                else:
                    print(" | d0.00", end="")
                if ground_truth is None:
                    print("")
                if reverse:
                    if fwd_PI > bwd_PI:
                        check = "✅" if ground_truth.has_edge(
                            feature, name) else "❌"
                        print(f" | {feature:4s} -> {name:4s} {check}")
                    else:
                        check = "✅" if ground_truth.has_edge(
                            name, feature) else "❌"
                        print(f" | {feature:4s} <- {name:4s} {check}")
                else:
                    print("")

    def _cluster_features(self, method, threshold):
        """
        Cluster the features of the data based on the linkage matrix obtained
        from the hierarchical clustering.

        Parameters
        ----------
        method : str
            Method to compute the correlation.
        threshold : float
            Threshold for the correlation.

        Returns
        -------
        clustered : pd.DataFrame
            Correlation matrix.
        sorted_colnames : List[str]
            List of sorted column names.
        """
        # Keep the indices to sort labels
        labels = fcluster(self.linkage_mat, threshold, criterion='distance')
        labels_order = np.argsort(labels)
        sorted_colnames = self.data.columns[labels_order]

        # Build a new dataframe with the sorted columns
        for idx, i in enumerate(sorted_colnames):
            if idx == 0:
                clustered = pd.DataFrame(self.data[i])
            else:
                df_to_append = pd.DataFrame(self.data[i])
                clustered = pd.concat([clustered, df_to_append], axis=1)
        return clustered.corr(method), sorted_colnames

    def hierarchical_dissimilarities(self):
        """
        Compute the dissimilarities between features in a hierarchical clustering.

        Returns
        -------
        hierarchical_dissimilarity : pd.DataFrame
            Dissimilarities between features.
        """
        feature_names = list(self.data.columns)
        hierarchical_dissimilarity = pd.DataFrame(columns=feature_names)
        clusters = self._clusters_from_linkage(self.linkage_mat, feature_names)
        for feature in feature_names:
            remaining_features = [f for f in feature_names if f != feature]
            for name in remaining_features:
                hierarchical_dissimilarity.loc[feature, name] = self._are_connected(
                    clusters, feature, name)

        hierarchical_dissimilarity[hierarchical_dissimilarity.isna()] = 0.0

        return hierarchical_dissimilarity

    def _clusters_from_linkage(self, linkage_mat: np.ndarray, features: List[str]):
        """
        Get the clusters from the linkage matrix.

        Parameters
        ----------
        linkage_mat : np.array
            The linkage matrix.
        features : List[str]
            The list of feature names.

        Returns
        -------
        clusters : dict
            The clusters from the linkage matrix.
        """
        clusters = {}
        Z = linkage_mat
        new_cluster_id = len(features)
        for n in range(Z.shape[0]):
            new_cluster = new_cluster_id
            new_cluster_id += 1
            feature_a = int(Z[n][0])
            feature_b = int(Z[n][1])
            if feature_a < len(features):
                feature_a = features[feature_a]
            else:
                feature_a = f'K{feature_a}'
            if feature_b < len(features):
                feature_b = features[feature_b]
            else:
                feature_b = f'K{feature_b}'
            clusters[f"K{new_cluster}"] = [(feature_a, feature_b), Z[n][2]]

        return clusters

    @staticmethod
    def _get_cluster(clusters: List[str], node: str) -> str:
        """
        Get the cluster that contains the given node.

        Parameters
        ----------
        clusters : List[str]
            The list of clusters.
        node : str
            The node to search for.

        Returns
        -------
        cluster : str
            The cluster that contains the given node.
        """
        for k in clusters.keys():
            if node in clusters[k][0]:
                return k
        return None

    @staticmethod
    def _is_cluster(node: str) -> bool:
        """
        Check if the given node is a cluster.

        Parameters
        ----------
        node : str
            The node to check.

        Returns
        -------
        is_cluster : bool
            True if the node is a cluster, False otherwise.
        """
        return node.startswith('K')

    @staticmethod
    def _contains_a_cluster(clusters: List[str], node: str) -> bool:
        """
        Check if the given node contains a cluster.

        Parameters
        ----------
        clusters : List[str]
            The list of clusters.
        node : str
            The node to check.

        Returns
        -------
        contains_cluster : bool
            True if the node contains a cluster, False otherwise.
        """
        return any([Hierarchies._is_cluster(n) for n in clusters[node][0]])

    @staticmethod
    def _get_cluster_element(clusters: List[str], node: str) -> str:
        """
        Get the cluster element from the given node.

        Parameters
        ----------
        clusters : List[str]
            The list of clusters.
        node : str
            The node to get the cluster element from.

        Returns
        -------
        cluster_element : str
            The cluster element from the given node.
        """
        for n in clusters[node][0]:
            if Hierarchies._is_cluster(n):
                return n
        return None

    @staticmethod
    def _in_cluster(cluster, node):
        """
        Check if the given node is in the given cluster.

        Parameters
        ----------
        cluster : str
            The cluster to check.
        node : str
            The node to check.

        Returns
        -------
        in_cluster : bool
            True if the node is in the cluster, False otherwise.
        """
        return node in cluster[0]

    @staticmethod
    def _in_same_cluster(clusters: List[str], node1: str, node2: str) -> str:
        """
        Check if the given nodes are in the same cluster.

        Parameters
        ----------
        clusters : List[str]
            The list of clusters.
        node1 : str
            The first node to check.
        node2 : str
            The second node to check.

        Returns
        -------
        cluster : str
            The cluster that contains both nodes, or None if they are not in the same cluster.
        """
        for k in clusters.keys():
            if node1 in clusters[k][0] and node2 in clusters[k][0]:
                return k
        return None

    def _are_connected(
            self,
            clusters: List[str],
            node1: str,
            node2: str) -> Tuple[bool, float]:
        """
        Determine if two nodes are connected in the hierarchical clustering represented
        in the clusters dictionary, obtained from the `linkage` function.

        Parameters
        ----------
            - clusters (List[str])
                List of clusters.
            - node1 (str)
                First node.
            - node2 (str)
                Second node.

        Returns
        -------
            - float
                Degree of disimilarity between the nodes. The higher the value, the more
                dissimilar the nodes are. If None is returned, the nodes are not connected.
        """
        k = Hierarchies._in_same_cluster(clusters, node1, node2)
        if k is not None:
            # Obtain the second valud in the dictionary of clusters for the pair of nodes
            return clusters[k][1]
        cluster1 = Hierarchies._get_cluster(clusters, node1)
        cluster2 = Hierarchies._get_cluster(clusters, node2)
        if Hierarchies._contains_a_cluster(clusters, cluster1):
            ref_cluster = Hierarchies._get_cluster_element(clusters, cluster1)
            if Hierarchies._in_cluster(clusters[ref_cluster], node2):
                disimilarity = clusters[Hierarchies._get_cluster(
                    clusters, node1)][1]
                return disimilarity
        if Hierarchies._contains_a_cluster(clusters, cluster2):
            ref_cluster = Hierarchies._get_cluster_element(clusters, cluster2)
            if Hierarchies._in_cluster(clusters[ref_cluster], node1):
                disimilarity = clusters[Hierarchies._get_cluster(
                    clusters, node2)][1]
                return disimilarity
        return None

def _get_directed_pair(g, u, v):
    if g.has_edge(u, v) and not g.has_edge(v, u):
        return (u, v)
    elif g.has_edge(v, u) and not g.has_edge(u, v):
        return (v, u)
    else:
        return None


def _get_direction(g: nx.DiGraph, u: str, v: str) -> Tuple[str, Union[int, None]]:
    """
    Returns the direction of the edge between u and v.

    Arguments:
    ----------
        - g (networkx.DiGraph): The graph to search in.
        - u (str): The source node.
        - v (str): The target node.

    Returns:
    --------
        - (str, int): A string respresenting the direction of the edge
            between u and v, and an integer representing it. The string
            can be either '-->', '<--', '<->' or ' · '. And the integer
            can be either 1, -1, 0 or None for each possibility.

    """

    if not g.has_node(u) or not g.has_node(v):
        raise ValueError(f"Node(s) {u} and/or {v} not in graph.")

    if g.has_edge(u, v) and not g.has_edge(v, u):
        return "-->", 1
    elif g.has_edge(v, u) and not g.has_edge(u, v):
        return "<--", -1
    elif g.has_edge(u, v) and g.has_edge(v, u):
        return "<->", 0
    else:
        return " · ", None


def connect_isolated_nodes(G, linkage_mat, feature_names, verbose=False):
    """
    Connect isolated nodes in the graph, based on their relationship in the
    hierarchical clustering provided through the linkage_mat.

    Arguments:
    ----------
        - G (networkx.DiGraph): The graph to search in.
        - linkage_mat (np.ndarray): The linkage matrix.
        - feature_names (List[str]): The list of feature names.
        - verbose (bool): Whether to print information about the process.

    Returns:
    --------
        - networkx.DiGraph: The graph with connected isolated nodes.

    Notes:
    ------
        The linkage matrix is a matrix of the following form:

            [i, j, distance, n_items, direction]

        where:
            - i, j: The indices of the two nodes in the linkage.
            - distance: The distance between the two nodes.
            - n_items: The number of items in the two nodes.
            - direction: The direction of the edge between the two nodes.

        and it can be obtained using the function: `compute_hierarchies`
    """
    node_names = [f for f in feature_names]
    G_h = nx.DiGraph()
    G_h.add_edges_from(G.edges(data=False))

    for i in range(linkage_mat.shape[0]):
        _, num = linkage_mat[i][2], int(linkage_mat[i][3])
        if num > 2:
            continue
        u, v = node_names[int(linkage_mat[i][0])
                          ], node_names[int(linkage_mat[i][1])]
        arrow, direction = _get_direction(G_h, u, v)

        # Consider only clusters formed by two nodes, not other clusters.
        features_cluster = True if num == 2 else False
        if direction is None and features_cluster:
            G_h.add_edge(u, v, weight=None)
            if verbose:
                print(f"Adding edge {u} {arrow} {v}")

    return G_h


def connect_hierarchies(G, linkage_mat, feature_names, verbose=False):
    cluster_id = len(feature_names)
    node_names = [f for f in feature_names]
    clusters = {}
    G_h = nx.DiGraph()
    G_h.add_edges_from(G.edges(data=True))
    if verbose:
        print(
            f"{'from':>4s}  :  {'to':<4s}  {'weight':6s}  {'n.items':7s} {'names':5s}")
    for i in range(linkage_mat.shape[0]):
        u, v = node_names[int(linkage_mat[i][0])
                          ], node_names[int(linkage_mat[i][1])]
        weight, num = linkage_mat[i][2], int(linkage_mat[i][3])
        arrow, _ = _get_direction(G_h, u, v)
        kname = f"K#{cluster_id}"
        node_names.append(kname)
        clusters[kname] = (u, v)
        if verbose:
            print(f"{u:>4s} {arrow:^s} {v:<4s}  {weight:6.4f}  {str(num):^7s} {kname:^5s}",
                  end="")
        cluster_id += 1

        # If I'm forming the first cluster, nothing to do.
        if cluster_id <= len(feature_names)+1:
            if verbose:
                print()
            continue

        # Determine if source is a node or a cluster
        if u in clusters:
            pair = _get_directed_pair(G_h, *clusters[u])
            if pair is not None:
                source = pair[1]
        else:
            source = u

        # Determine if target is a node or cluster. If node, continue
        if v in clusters:
            pair = _get_directed_pair(G_h, *clusters[v])
            if pair is not None:
                target = pair[0]
            else:
                if verbose:
                    print()
                continue
        else:
            if verbose:
                print()
            continue

        if _get_directed_pair(G_h, source, target) is None:
            if verbose:
                print(f" Add {source} --> {target}")
            G_h.add_edge(source, target)
        else:
            if verbose:
                print()
            continue

    return G_h


def plot_dendogram_correlations(correlations, feature_names: List[str], **kwargs):
    """
    Plot the dendrogram of the correlation matrix.

    Parameters
    ----------
        - correlations (pd.DataFrame)
            Correlation matrix.
        - feature_names (List[str])
            List of feature names.
        - kwargs
            Keyword arguments to be passed to the plot_dendogram function.
    """
    figsize = kwargs.get("figsize", (5, 3))
    dissimilarity = 1 - abs(correlations)
    Z = linkage(squareform(dissimilarity), 'single')
    plt.figure(figsize=figsize)
    ax = plt.gca()
    dendrogram(Z, labels=feature_names, orientation='left', ax=ax)
    plt.show()
    return Z


if __name__ == "__main__":
    m_alpha = 0.8
    m_c = 15

    test_data = pd.read_csv("/Users/renero/phd/data/generated_linear_10.csv")
    h = Hierarchies(method='mic', mic_alpha=m_alpha, mic_c=m_c)
    h.fit(test_data)
    h.plot()
    plt.show()



import numpy as np
import pytest

from causalexplain.estimators.cam import computeScoreMat as compute_score_mat_module


def test_compute_score_mat_invokes_parallel_and_shapes(monkeypatch):
    calls = []

    def fake_parallel(row_parents, score_name, X, sel_mat, verbose, node2, i, pars_score, interv_mat, interv_data):
        calls.append((tuple(row_parents.flatten()), score_name, node2, i, interv_data))
        return float(node2 * 10 + i)

    monkeypatch.setattr(compute_score_mat_module, "computeScoreMatParallel", fake_parallel)

    X = np.array([[1.0, 2.0], [3.0, 4.0]])
    sel_mat = np.ones((2, 2), dtype=bool)
    result = compute_score_mat_module.computeScoreMat(
        X=X,
        score_name="test",
        num_parents=1,
        verbose=False,
        num_cores=1,
        sel_mat=sel_mat,
        pars_score={},
        interv_mat=None,
        interv_data=False,
    )

    assert set(result) == {"scoreMat", "rowParents", "scoreEmptyNodes"}
    assert result["scoreMat"].shape == (2, 2)
    assert calls[0][2:] == (0, 0, False)
    assert np.array_equal(result["scoreMat"], np.array([[0.0, 1.0], [10.0, 11.0]]))


def test_compute_score_mat_uses_interventions(monkeypatch):
    monkeypatch.setattr(
        compute_score_mat_module,
        "computeScoreMatParallel",
        lambda *args, **kwargs: 1.0,
    )
    X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
    interv_mat = np.array([[True, False], [False, False], [False, False]])
    sel_mat = np.ones((2, 2), dtype=bool)

    result = compute_score_mat_module.computeScoreMat(
        X=X,
        score_name="test",
        num_parents=1,
        verbose=False,
        num_cores=1,
        sel_mat=sel_mat,
        pars_score={},
        interv_mat=interv_mat,
        interv_data=True,
    )

    assert result["scoreEmptyNodes"][0] == pytest.approx(-np.log(1))
    assert result["scoreMat"][0, 1] == pytest.approx(1 - (-np.log(np.var(X[:, 1]))))



"""

(C) Original Code from R implementation of the Causal Additive Model (CAM)

@article{buhlmann2014cam,
  title={CAM: Causal additive models, high-dimensional order search and 
          penalized regression},
  author={B{\"u}hlmann, Peter and Peters, Jonas and Ernest, Jan},
  journal={The Annals of Statistics},
  volume={42},
  number={6},
  pages={2526--2556},
  year={2014},
  publisher={Institute of Mathematical Statistics}
}

- **Imports**: Imported necessary modules and functions. Assumed that
`computeScoreMat`, `updateScoreMat`, `pruning`, `selGamBoost`, and `selGam`
are defined in separate Python files in the same directory.
- **Function Definition**: Translated the R function `CAM` to Python.
- **Variable Initialization**: Initialized variables and handled default values.
- **Variable Selection**: Used `numpy` and `multiprocessing` for parallel 
    processing.
- **Edge Inclusion**: Translated the logic for including edges and updating the
score matrix.
- **Pruning**: Translated the pruning step.
- **Output and Return**: Collected and printed the results.

Make sure the corresponding Python files (`computeScoreMat.py`, 
    `updateScoreMat.py`,
`pruning.py`, `selGamBoost.py`, `selGam.py`) are present in the same directory 
and contain the necessary functions.
"""
# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init
# pylint: disable=C0103:invalid-name, W0221:arguments-differ
# pylint: disable=C0116:missing-function-docstring, W0511:fixme
# pylint: disable=R0913:too-many-arguments, E0401:import-error
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

import os
import time
from multiprocessing import Pool

import networkx as nx
import numpy as np
import pandas as pd

from ...common import utils
from .computeScoreMat import computeScoreMat
from .pruning import pruning
from .selGam import selGam
from .selGamBoost import selGamBoost
from .updateScoreMat import updateScoreMat
from ...metrics.compare_graphs import evaluate_graph


class CAM:
    def __init__(
            self,
            name:str,
            scoreName="SEMGAM",
            parsScore=None,
            numCores=1,
            maxNumParents=None,
            verbose=False,
            variableSel=False,
            variableSelMethod=selGamBoost,
            variableSelMethodPars=None,
            pruning=True,
            pruneMethod=selGam,
            pruneMethodPars={"cutOffPVal": 0.05, "numBasisFcts": 10},
            intervData=False,
            intervMat=None):

        self.name = name
        self.scoreName = scoreName
        self.parsScore = parsScore
        self.numCores = numCores
        self.maxNumParents = maxNumParents
        self.verbose = verbose
        self.variableSel = variableSel
        self.variableSelMethod = variableSelMethod
        self.variableSelMethodPars = variableSelMethodPars
        self.pruning = pruning
        self.pruneMethod = pruneMethod
        self.pruneMethodPars = pruneMethodPars
        self.intervData = intervData
        self.intervMat = intervMat

        self.is_fitted_ = False
        self.metrics = None
        self.dag = None

        if parsScore is None:
            self.parsScore = {"numBasisFcts": 10}
        if variableSelMethodPars is None:
            self.variableSelMethodPars = {
                "atLeastThatMuchSelected": 0.02, "atMostThatManyNeighbors": 10}
        if pruneMethodPars is None:
            self.pruneMethodPars = {"cutOffPVal": 0.001, "numBasisFcts": 10}

        self.maxNumParents = maxNumParents

    def fit(self, X):
        """
        This method implements the entire CAM algorithm. Translated from the R code.

        Parameters
        ----------
        X : np.array
            Observational data

        Returns
        -------
        edgeList : list
            List of edges
        scoreVec : list
            List of scores
        """
        if self.verbose:
            print(f"number of cores: {self.numCores}")

        if self.maxNumParents is None:
            self.maxNumParents = min(X.shape[1] - 1, round(X.shape[0] / 20))

        timeCycle = 0
        timeUpdate = 0
        timeScoreMat = 0
        timeSel = 0
        timePrune = 0
        timeMax = 0

        scoreVec = []
        edgeList = []

        counterUpdate = 0

        self.feature_names = list(X.columns)
        X = X.values
        p = X.shape[1]

        if self.variableSel:
            start_time = time.time()
            if self.intervData:
                X2 = X[np.sum(self.intervMat, axis=1) == 0, :]
                if self.verbose:
                    print("The preliminary neighbourhood selection is done with the " +
                          "observational data only.")
            else:
                X2 = X

            if self.numCores == 1:
                selMat = np.array(
                    [self.variableSelMethod(X2, self.variableSelMethodPars, self.verbose) for _ in range(p)])
            else:
                with Pool(self.numCores) as pool:
                    selMat = np.array(pool.starmap(self.variableSelMethod, [
                        (X2, self.variableSelMethodPars, self.verbose)
                        for _ in range(p)]))

            cou = sum(2 ** np.sum(selMat[:, jk]) for jk in range(p))
            if self.verbose:
                print(
                    f"Instead of p2^(p-1) -Sillander- {p * 2 ** (p - 1)} we have {cou}")
                print(
                    f"Greedy, on the other hand, is computing {np.sum(selMat)} entries.")

            timeSel += time.time() - start_time
        else:
            selMat = np.ones((p, p), dtype=bool)
            print('NO variable selection') if self.verbose else None

        if self.variableSel and self.verbose:
            if p < 30:
                print("This is the matrix of possible parents after the first step.")
                print(selMat)

        start_time = time.time()
        computeScoreMatTmp = computeScoreMat(
            X, score_name=self.scoreName, num_parents=1, num_cores=self.numCores,
            verbose=self.verbose, sel_mat=selMat, pars_score=self.parsScore, interv_mat=self.intervMat,
            interv_data=self.intervData)
        timeScoreMat += time.time() - start_time

        pathMatrix = np.eye(p, dtype=int)
        self.adj = np.zeros((p, p), dtype=int)
        scoreNodes = computeScoreMatTmp['scoreEmptyNodes']

        if self.verbose:
            print(f"scoreNodes: {scoreNodes}")
            print("Contents of computeScoreMat@CAM()")
            for i in range(computeScoreMatTmp["scoreMat"].shape[0]):
                for j in range(computeScoreMatTmp["scoreMat"].shape[1]):
                    if computeScoreMatTmp['scoreMat'][i, j] == -np.inf:
                        print(" -inf   ", end="")
                    else:
                        print(
                            f"{computeScoreMatTmp['scoreMat'][i, j]:+.4f} ", end="")
                print("")

            print("Pre-WHILE condition: ", np.sum(computeScoreMatTmp['scoreMat'] != -np.Inf))

        while np.sum(computeScoreMatTmp['scoreMat'] != -np.inf) > 0:
            start_time = time.time()
            ix_max = np.unravel_index(np.argmax(
                computeScoreMatTmp['scoreMat']), computeScoreMatTmp['scoreMat'].shape)

            timeMax += time.time() - start_time
            self.adj[ix_max] = 1
            scoreNodes[ix_max[1]] += computeScoreMatTmp['scoreMat'][ix_max]

            if self.verbose:
                print(f"ix_max: {ix_max}")
                print(f"\nIncluded edge (from, to) {ix_max}")

            computeScoreMatTmp['scoreMat'][ix_max] = -np.inf

            start_time = time.time()
            pathMatrix[ix_max[0], ix_max[1]] = 1
            DescOfNewChild = np.where(pathMatrix[ix_max[1], :] == 1)[0]
            AncOfNewParent = np.where(pathMatrix[:, ix_max[0]] == 1)[0]
            pathMatrix[np.ix_(AncOfNewParent, DescOfNewChild)] = 1
            computeScoreMatTmp['scoreMat'][pathMatrix.T == 1] = -np.inf
            computeScoreMatTmp['scoreMat'][ix_max[1], ix_max[0]] = -np.inf
            timeCycle += time.time() - start_time

            scoreVec.append(np.sum(scoreNodes))
            edgeList.append(ix_max)

            start_time = time.time()
            computeScoreMatTmp['scoreMat'] = updateScoreMat(
                computeScoreMatTmp['scoreMat'], X, score_name=self.scoreName, i=ix_max[0],
                j=ix_max[1], score_nodes=scoreNodes, adj=self.adj, num_cores=self.numCores,
                verbose=self.verbose, max_num_parents=self.maxNumParents, pars_score=self.parsScore,
                interv_mat=self.intervMat, interv_data=self.intervData)
            timeUpdate += time.time() - start_time

            counterUpdate += 1

        if self.verbose:
            print("\n--------------------------")
            print("Finished step 2 ----------")
            print("--------------------------")
        # Step 3

        if self.pruning:
            if self.intervData:
                X2 = X[np.sum(self.intervMat, axis=1) == 0, :]
                print("The preliminary neighbourhood selection is done with the " +
                      "observational data only.")
            else:
                X2 = X

            if self.verbose:
                print("\n Performing pruning ... \n")

            start_time = time.time()
            self.adj = pruning(X=X2, G=self.adj, prune_method=self.pruneMethod,
                          prune_method_pars=self.pruneMethodPars, verbose=self.verbose)
            timePrune += time.time() - start_time

        timeTotal = timeSel + timeScoreMat + timeCycle + timeUpdate + timeMax + timePrune
        if self.verbose:
            print("\nTiming:")
            print(f"  Time for variable selection: {timeSel:.2f}")
            print(f"  Time computing the initial scoreMat: {timeScoreMat:.2f}")
            print(f"  Time checking for cycles: {timeCycle:.2f}")
            print(
                f"  Time computing updates for the scoreMat: {timeUpdate:.2f}, "
                f"doing {counterUpdate} updates.")
            print(f"  Time for pruning: {timePrune:.2f}")
            print(f"  Time for finding maximum: {timeMax:.2f}")
            print(f"  Time in total: {timeTotal:.2f}")

        result = {
            "Adj": self.adj,
            "Score": np.sum(scoreNodes),
            "timesVec":
                [timeSel, timeScoreMat, timeCycle, timeUpdate,
                    timePrune, timeMax, timeTotal],
            "scoreVec": scoreVec,
            "edgeList": edgeList
        }

        self.is_fitted_ = True
        return self

    def predict(self, ref_graph: nx.DiGraph = None):
        if not self.is_fitted_:
            raise ValueError(
                f"This {self.__class__.__name__} instance is not fitted yet."
                f"Call 'fit' with appropriate arguments before using this method.")

        self.dag = utils.graph_from_adjacency(self.adj, self.feature_names)
        if ref_graph is not None:
            self.metrics = evaluate_graph(
                ref_graph, self.dag, self.feature_names)
        else:
            self.metrics = None

        return self.dag

    def fit_predict(
            self,
            train_data:pd.DataFrame,
            test_data:pd.DataFrame = None,
            ref_graph: nx.DiGraph = None):
        self.fit(train_data)
        self.predict(ref_graph)

        return self.dag


def main(dataset_name,
         input_path="/Users/renero/phd/data/sachs",
         output_path="/Users/renero/phd/output/",
         save=False):

    data = pd.read_csv(os.path.join(input_path, dataset_name) + ".csv")
    ref_graph = utils.graph_from_dot_file(
        os.path.join(input_path, dataset_name) + ".dot")

    cam = CAM(name="main_run",
              pruning=True,
              pruneMethodPars={"cutOffPVal": 0.05, "numBasisFcts": 10},
              verbose=False)
    cam.fit_predict(data, ref_graph=ref_graph)
    print(cam.metrics)


if __name__ == "__main__":
    main("sachs")
